import enum

class SemanticVariant(enum.Enum):
    Primary = "primary"
    Secondary = "secondary"
    Danger = "danger"
    Warning = "warning"
    Info = "info"
    Dark = "dark"
    Light = "light"
    Body = "body"
    Muted = "muted"

    def __str__(self):
        return self.value

class SemanticClass(enum.Enum):
    Active='active'
    Alert = 'alert'
    AlertDanger = 'alert-danger'
    AlertDark = 'alert-dark'
    AlertDismissible = 'alert-dismissible'
    AlertHeading = 'alert-heading'
    AlertInfo = 'alert-info'
    AlertLight = 'alert-light'
    AlertLink = 'alert-link'
    AlertPrimary = 'alert-primary'
    AlertSecondary = 'alert-secondary'
    AlertSuccess = 'alert-success'
    AlertWarning = 'alert-warning'
    AlignBaseline = 'align-baseline'
    AlignBottom = 'align-bottom'
    AlignContentAround = 'align-content-around'
    AlignContentBetween = 'align-content-between'
    AlignContentCenter = 'align-content-center'
    AlignContentEnd = 'align-content-end'
    AlignContentLargeAround = 'align-content-lg-around'
    AlignContentLargeBetween = 'align-content-lg-between'
    AlignContentLargeCenter = 'align-content-lg-center'
    AlignContentLargeEnd = 'align-content-lg-end'
    AlignContentLargeStart = 'align-content-lg-start'
    AlignContentLargeStretch = 'align-content-lg-stretch'
    AlignContentMediumAround = 'align-content-md-around'
    AlignContentMediumBetween = 'align-content-md-between'
    AlignContentMediumCenter = 'align-content-md-center'
    AlignContentMediumEnd = 'align-content-md-end'
    AlignContentMediumStart = 'align-content-md-start'
    AlignContentMediumStretch = 'align-content-md-stretch'
    AlignContentSmallAround = 'align-content-sm-around'
    AlignContentSmallBetween = 'align-content-sm-between'
    AlignContentSmallCenter = 'align-content-sm-center'
    AlignContentSmallEnd = 'align-content-sm-end'
    AlignContentSmallStart = 'align-content-sm-start'
    AlignContentSmallStretch = 'align-content-sm-stretch'
    AlignContentStart = 'align-content-start'
    AlignContentStretch = 'align-content-stretch'
    AlignContentExtraLargeAround = 'align-content-xl-around'
    AlignContentExtraLargeBetween = 'align-content-xl-between'
    AlignContentExtraLargeCenter = 'align-content-xl-center'
    AlignContentExtraLargeEnd = 'align-content-xl-end'
    AlignContentExtraLargeStart = 'align-content-xl-start'
    AlignContentExtraLargeStretch = 'align-content-xl-stretch'
    AlignItemsBaseline = 'align-items-baseline'
    AlignItemsCenter = 'align-items-center'
    AlignItemsEnd = 'align-items-end'
    AlignItemsLargeBaseline = 'align-items-lg-baseline'
    AlignItemsLargeCenter = 'align-items-lg-center'
    AlignItemsLargeEnd = 'align-items-lg-end'
    AlignItemsLargeStart = 'align-items-lg-start'
    AlignItemsLargeStretch = 'align-items-lg-stretch'
    AlignItemsMediumBaseline = 'align-items-md-baseline'
    AlignItemsMediumCenter = 'align-items-md-center'
    AlignItemsMediumEnd = 'align-items-md-end'
    AlignItemsMediumStart = 'align-items-md-start'
    AlignItemsMediumStretch = 'align-items-md-stretch'
    AlignItemsSmallBaseline = 'align-items-sm-baseline'
    AlignItemsSmallCenter = 'align-items-sm-center'
    AlignItemsSmallEnd = 'align-items-sm-end'
    AlignItemsSmallStart = 'align-items-sm-start'
    AlignItemsSmallStretch = 'align-items-sm-stretch'
    AlignItemsStart = 'align-items-start'
    AlignItemsStretch = 'align-items-stretch'
    AlignItemsExtraLargeBaseline = 'align-items-xl-baseline'
    AlignItemsExtraLargeCenter = 'align-items-xl-center'
    AlignItemsExtraLargeEnd = 'align-items-xl-end'
    AlignItemsExtraLargeStart = 'align-items-xl-start'
    AlignItemsExtraLargeStretch = 'align-items-xl-stretch'
    AlignMiddle = 'align-middle'
    AlignSelfAuto = 'align-self-auto'
    AlignSelfBaseline = 'align-self-baseline'
    AlignSelfCenter = 'align-self-center'
    AlignSelfEnd = 'align-self-end'
    AlignSelfLargeAuto = 'align-self-lg-auto'
    AlignSelfLargeBaseline = 'align-self-lg-baseline'
    AlignSelfLargeCenter = 'align-self-lg-center'
    AlignSelfLargeEnd = 'align-self-lg-end'
    AlignSelfLargeStart = 'align-self-lg-start'
    AlignSelfLargeStretch = 'align-self-lg-stretch'
    AlignSelfMediumAuto = 'align-self-md-auto'
    AlignSelfMediumBaseline = 'align-self-md-baseline'
    AlignSelfMediumCenter = 'align-self-md-center'
    AlignSelfMediumEnd = 'align-self-md-end'
    AlignSelfMediumStart = 'align-self-md-start'
    AlignSelfMediumStretch = 'align-self-md-stretch'
    AlignSelfSmallAuto = 'align-self-sm-auto'
    AlignSelfSmallBaseline = 'align-self-sm-baseline'
    AlignSelfSmallCenter = 'align-self-sm-center'
    AlignSelfSmallEnd = 'align-self-sm-end'
    AlignSelfSmallStart = 'align-self-sm-start'
    AlignSelfSmallStretch = 'align-self-sm-stretch'
    AlignSelfStart = 'align-self-start'
    AlignSelfStretch = 'align-self-stretch'
    AlignSelfExtraLargeAuto = 'align-self-xl-auto'
    AlignSelfExtraLargeBaseline = 'align-self-xl-baseline'
    AlignSelfExtraLargeCenter = 'align-self-xl-center'
    AlignSelfExtraLargeEnd = 'align-self-xl-end'
    AlignSelfExtraLargeStart = 'align-self-xl-start'
    AlignSelfExtraLargeStretch = 'align-self-xl-stretch'
    AlignTextBottom = 'align-text-bottom'
    AlignTextTop = 'align-text-top'
    AlignTop = 'align-top'
    Arrow = 'arrow'
    Badge = 'badge'
    BadgeDanger = 'badge-danger'
    BadgeDark = 'badge-dark'
    BadgeInfo = 'badge-info'
    BadgeLight = 'badge-light'
    BadgePill = 'badge-pill'
    BadgePrimary = 'badge-primary'
    BadgeSecondary = 'badge-secondary'
    BadgeSuccess = 'badge-success'
    BadgeWarning = 'badge-warning'
    BackgroundDanger = 'bg-danger'
    BackgroundDark = 'bg-dark'
    BackgroundInfo = 'bg-info'
    BackgroundLight = 'bg-light'
    BackgroundPrimary = 'bg-primary'
    BackgroundSecondary = 'bg-secondary'
    BackgroundSuccess = 'bg-success'
    BackgroundTransparent = 'bg-transparent'
    BackgroundWarning = 'bg-warning'
    BackgroundWhite = 'bg-white'
    Blockquote = 'blockquote'
    BlockquoteFooter = 'blockquote-footer'
    Border = 'border'
    Border0 = 'border-0'
    BorderBottom = 'border-bottom'
    BorderBottom0 = 'border-bottom-0'
    BorderDanger = 'border-danger'
    BorderDark = 'border-dark'
    BorderInfo = 'border-info'
    BorderLeft = 'border-left'
    BorderLeft0 = 'border-left-0'
    BorderLight = 'border-light'
    BorderPrimary = 'border-primary'
    BorderRight = 'border-right'
    BorderRight0 = 'border-right-0'
    BorderSecondary = 'border-secondary'
    BorderSuccess = 'border-success'
    BorderTop = 'border-top'
    BorderTop0 = 'border-top-0'
    BorderWarning = 'border-warning'
    BorderWhite = 'border-white'
    Breadcrumb = 'breadcrumb'
    BreadcrumbItem = 'breadcrumb-item'
    BsPopoverBottom = 'bs-popover-bottom'
    BsPopoverLeft = 'bs-popover-left'
    BsPopoverRight = 'bs-popover-right'
    BsPopoverTop = 'bs-popover-top'
    BsTooltipBottom = 'bs-tooltip-bottom'
    BsTooltipLeft = 'bs-tooltip-left'
    BsTooltipRight = 'bs-tooltip-right'
    BsTooltipTop = 'bs-tooltip-top'
    Button = 'btn'
    ButtonBlock = 'btn-block'
    ButtonDanger = 'btn-danger'
    ButtonDark = 'btn-dark'
    ButtonGroup = 'btn-group'
    ButtonGroupVertical = 'btn-group-vertical'
    ButtonInfo = 'btn-info'
    ButtonLarge = 'btn-lg'
    ButtonLight = 'btn-light'
    ButtonLink = 'btn-link'
    ButtonOutlineDanger = 'btn-outline-danger'
    ButtonOutlineDark = 'btn-outline-dark'
    ButtonOutlineInfo = 'btn-outline-info'
    ButtonOutlineLight = 'btn-outline-light'
    ButtonOutlinePrimary = 'btn-outline-primary'
    ButtonOutlineSecondary = 'btn-outline-secondary'
    ButtonOutlineSuccess = 'btn-outline-success'
    ButtonOutlineWarning = 'btn-outline-warning'
    ButtonPrimary = 'btn-primary'
    ButtonSecondary = 'btn-secondary'
    ButtonSmall = 'btn-sm'
    ButtonSuccess = 'btn-success'
    ButtonToolbar = 'btn-toolbar'
    ButtonWarning = 'btn-warning'
    Card = 'card'
    CardBody = 'card-body'
    CardColumns = 'card-columns'
    CardDeck = 'card-deck'
    CardFooter = 'card-footer'
    CardGroup = 'card-group'
    CardHeader = 'card-header'
    CardHeaderPills = 'card-header-pills'
    CardHeaderTabs = 'card-header-tabs'
    CardImage = 'card-img'
    CardImageBottom = 'card-img-bottom'
    CardImageOverlay = 'card-img-overlay'
    CardImageTop = 'card-img-top'
    CardLink = 'card-link'
    CardSubtitle = 'card-subtitle'
    CardTitle = 'card-title'
    Carousel = 'carousel'
    CarouselCaption = 'carousel-caption'
    CarouselControlNext = 'carousel-control-next'
    CarouselControlNextIcon = 'carousel-control-next-icon'
    CarouselControlPrev = 'carousel-control-prev'
    CarouselControlPrevIcon = 'carousel-control-prev-icon'
    CarouselIndicators = 'carousel-indicators'
    CarouselInner = 'carousel-inner'
    CarouselItem = 'carousel-item'
    CarouselItemLeft = 'carousel-item-left'
    CarouselItemNext = 'carousel-item-next'
    CarouselItemPrev = 'carousel-item-prev'
    CarouselItemRight = 'carousel-item-right'
    Close = 'close'
    Column = 'col'
    Column1 = 'col-1'
    Column10 = 'col-10'
    Column11 = 'col-11'
    Column12 = 'col-12'
    Column2 = 'col-2'
    Column3 = 'col-3'
    Column4 = 'col-4'
    Column5 = 'col-5'
    Column6 = 'col-6'
    Column7 = 'col-7'
    Column8 = 'col-8'
    Column9 = 'col-9'
    ColumnAuto = 'col-auto'
    ColumnFormLabel = 'col-form-label'
    ColumnFormLabelLarge = 'col-form-label-lg'
    ColumnFormLabelSmall = 'col-form-label-sm'
    Collapsing = 'collapsing'
    ColumnLarge = 'col-lg'
    ColumnLarge1 = 'col-lg-1'
    ColumnLarge10 = 'col-lg-10'
    ColumnLarge11 = 'col-lg-11'
    ColumnLarge12 = 'col-lg-12'
    ColumnLarge2 = 'col-lg-2'
    ColumnLarge3 = 'col-lg-3'
    ColumnLarge4 = 'col-lg-4'
    ColumnLarge5 = 'col-lg-5'
    ColumnLarge6 = 'col-lg-6'
    ColumnLarge7 = 'col-lg-7'
    ColumnLarge8 = 'col-lg-8'
    ColumnLarge9 = 'col-lg-9'
    ColumnLargeAuto = 'col-lg-auto'
    ColumnMedium = 'col-md'
    ColumnMedium1 = 'col-md-1'
    ColumnMedium10 = 'col-md-10'
    ColumnMedium11 = 'col-md-11'
    ColumnMedium12 = 'col-md-12'
    ColumnMedium2 = 'col-md-2'
    ColumnMedium3 = 'col-md-3'
    ColumnMedium4 = 'col-md-4'
    ColumnMedium5 = 'col-md-5'
    ColumnMedium6 = 'col-md-6'
    ColumnMedium7 = 'col-md-7'
    ColumnMedium8 = 'col-md-8'
    ColumnMedium9 = 'col-md-9'
    ColumnMediumAuto = 'col-md-auto'
    ColumnSmall = 'col-sm'
    ColumnSmall1 = 'col-sm-1'
    ColumnSmall10 = 'col-sm-10'
    ColumnSmall11 = 'col-sm-11'
    ColumnSmall12 = 'col-sm-12'
    ColumnSmall2 = 'col-sm-2'
    ColumnSmall3 = 'col-sm-3'
    ColumnSmall4 = 'col-sm-4'
    ColumnSmall5 = 'col-sm-5'
    ColumnSmall6 = 'col-sm-6'
    ColumnSmall7 = 'col-sm-7'
    ColumnSmall8 = 'col-sm-8'
    ColumnSmall9 = 'col-sm-9'
    ColumnSmallAuto = 'col-sm-auto'
    ColumnExtraLarge = 'col-xl'
    ColumnExtraLarge1 = 'col-xl-1'
    ColumnExtraLarge10 = 'col-xl-10'
    ColumnExtraLarge11 = 'col-xl-11'
    ColumnExtraLarge12 = 'col-xl-12'
    ColumnExtraLarge2 = 'col-xl-2'
    ColumnExtraLarge3 = 'col-xl-3'
    ColumnExtraLarge4 = 'col-xl-4'
    ColumnExtraLarge5 = 'col-xl-5'
    ColumnExtraLarge6 = 'col-xl-6'
    ColumnExtraLarge7 = 'col-xl-7'
    ColumnExtraLarge8 = 'col-xl-8'
    ColumnExtraLarge9 = 'col-xl-9'
    ColumnExtraLargeAuto = 'col-xl-auto'
    Container = 'container'
    ContainerFluid = 'container-fluid'
    CustomControl = 'custom-control'
    CustomControlInline = 'custom-control-inline'
    CustomControlInput = 'custom-control-input'
    CustomControlLabel = 'custom-control-label'
    CustomFile = 'custom-file'
    CustomFileInput = 'custom-file-input'
    CustomFileLabel = 'custom-file-label'
    CustomRange = 'custom-range'
    CustomSelect = 'custom-select'
    CustomSelectLarge = 'custom-select-lg'
    CustomSelectSmall = 'custom-select-sm'
    CustomSwitch = 'custom-switch'
    DisplayBlock = 'd-block'
    DisplayFlex = 'd-flex'
    DisplayInline = 'd-inline'
    DisplayInlineBlock = 'd-inline-block'
    DisplayInlineFlex = 'd-inline-flex'
    Disabled = 'disabled'
    Display1 = 'display-1'
    Display2 = 'display-2'
    Display3 = 'display-3'
    Display4 = 'display-4'
    DisplayLargeBlock = 'd-lg-block'
    DisplayLargeFlex = 'd-lg-flex'
    DisplayLargeInline = 'd-lg-inline'
    DisplayLargeInlineBlock = 'd-lg-inline-block'
    DisplayLargeInlineFlex = 'd-lg-inline-flex'
    DisplayLargeNone = 'd-lg-none'
    DisplayLargeTable = 'd-lg-table'
    DisplayLargeTableCell = 'd-lg-table-cell'
    DisplayLargeTableRow = 'd-lg-table-row'
    DisplayMediumBlock = 'd-md-block'
    DisplayMediumFlex = 'd-md-flex'
    DisplayMediumInline = 'd-md-inline'
    DisplayMediumInlineBlock = 'd-md-inline-block'
    DisplayMediumInlineFlex = 'd-md-inline-flex'
    DisplayMediumNone = 'd-md-none'
    DisplayMediumTable = 'd-md-table'
    DisplayMediumTableCell = 'd-md-table-cell'
    DisplayMediumTableRow = 'd-md-table-row'
    DisplayNone = 'd-none'
    DisplayPrintBlock = 'd-print-block'
    DisplayPrintFlex = 'd-print-flex'
    DisplayPrintInline = 'd-print-inline'
    DisplayPrintInlineBlock = 'd-print-inline-block'
    DisplayPrintInlineFlex = 'd-print-inline-flex'
    DisplayPrintNone = 'd-print-none'
    DisplayPrintTable = 'd-print-table'
    DisplayPrintTableCell = 'd-print-table-cell'
    DisplayPrintTableRow = 'd-print-table-row'
    Dropdown = 'dropdown'
    DropdownDivider = 'dropdown-divider'
    DropdownHeader = 'dropdown-header'
    DropdownItem = 'dropdown-item'
    DropdownItemText = 'dropdown-item-text'
    DropdownMenu = 'dropdown-menu'
    DropdownMenuLeft = 'dropdown-menu-left'
    DropdownMenuLargeLeft = 'dropdown-menu-lg-left'
    DropdownMenuLargeRight = 'dropdown-menu-lg-right'
    DropdownMenuMediumLeft = 'dropdown-menu-md-left'
    DropdownMenuMediumRight = 'dropdown-menu-md-right'
    DropdownMenuRight = 'dropdown-menu-right'
    DropdownMenuSmallLeft = 'dropdown-menu-sm-left'
    DropdownMenuSmallRight = 'dropdown-menu-sm-right'
    DropdownMenuExtraLargeLeft = 'dropdown-menu-xl-left'
    DropdownMenuExtraLargeRight = 'dropdown-menu-xl-right'
    DropdownToggle = 'dropdown-toggle'
    DropdownToggleSplit = 'dropdown-toggle-split'
    Dropleft = 'dropleft'
    Dropright = 'dropright'
    Dropup = 'dropup'
    DisplaySmallBlock = 'd-sm-block'
    DisplaySmallFlex = 'd-sm-flex'
    DisplaySmallInline = 'd-sm-inline'
    DisplaySmallInlineBlock = 'd-sm-inline-block'
    DisplaySmallInlineFlex = 'd-sm-inline-flex'
    DisplaySmallNone = 'd-sm-none'
    DisplaySmallTable = 'd-sm-table'
    DisplaySmallTableCell = 'd-sm-table-cell'
    DisplaySmallTableRow = 'd-sm-table-row'
    DisplayTable = 'd-table'
    DisplayTableCell = 'd-table-cell'
    DisplayTableRow = 'd-table-row'
    DisplayExtraLargeBlock = 'd-xl-block'
    DisplayExtraLargeFlex = 'd-xl-flex'
    DisplayExtraLargeInline = 'd-xl-inline'
    DisplayExtraLargeInlineBlock = 'd-xl-inline-block'
    DisplayExtraLargeInlineFlex = 'd-xl-inline-flex'
    DisplayExtraLargeNone = 'd-xl-none'
    DisplayExtraLargeTable = 'd-xl-table'
    DisplayExtraLargeTableCell = 'd-xl-table-cell'
    DisplayExtraLargeTableRow = 'd-xl-table-row'
    EmbedResponsive = 'embed-responsive'
    EmbedResponsiveItem = 'embed-responsive-item'
    Fade = 'fade'
    Figure = 'figure'
    FigureCaption = 'figure-caption'
    FigureImage = 'figure-img'
    FixedBottom = 'fixed-bottom'
    FixedTop = 'fixed-top'
    FlexColumn = 'flex-column'
    FlexColumnReverse = 'flex-column-reverse'
    FlexFill = 'flex-fill'
    FlexGrow0 = 'flex-grow-0'
    FlexGrow1 = 'flex-grow-1'
    FlexLargeColumn = 'flex-lg-column'
    FlexLargeColumnReverse = 'flex-lg-column-reverse'
    FlexLargeFill = 'flex-lg-fill'
    FlexLargeGrow0 = 'flex-lg-grow-0'
    FlexLargeGrow1 = 'flex-lg-grow-1'
    FlexLargeNowrap = 'flex-lg-nowrap'
    FlexLargeRow = 'flex-lg-row'
    FlexLargeRowReverse = 'flex-lg-row-reverse'
    FlexLargeShrink0 = 'flex-lg-shrink-0'
    FlexLargeShrink1 = 'flex-lg-shrink-1'
    FlexLargeWrap = 'flex-lg-wrap'
    FlexLargeWrapReverse = 'flex-lg-wrap-reverse'
    FlexMediumColumn = 'flex-md-column'
    FlexMediumColumnReverse = 'flex-md-column-reverse'
    FlexMediumFill = 'flex-md-fill'
    FlexMediumGrow0 = 'flex-md-grow-0'
    FlexMediumGrow1 = 'flex-md-grow-1'
    FlexMediumNowrap = 'flex-md-nowrap'
    FlexMediumRow = 'flex-md-row'
    FlexMediumRowReverse = 'flex-md-row-reverse'
    FlexMediumShrink0 = 'flex-md-shrink-0'
    FlexMediumShrink1 = 'flex-md-shrink-1'
    FlexMediumWrap = 'flex-md-wrap'
    FlexMediumWrapReverse = 'flex-md-wrap-reverse'
    FlexNowrap = 'flex-nowrap'
    FlexRow = 'flex-row'
    FlexRowReverse = 'flex-row-reverse'
    FlexShrink0 = 'flex-shrink-0'
    FlexShrink1 = 'flex-shrink-1'
    FlexSmallColumn = 'flex-sm-column'
    FlexSmallColumnReverse = 'flex-sm-column-reverse'
    FlexSmallFill = 'flex-sm-fill'
    FlexSmallGrow0 = 'flex-sm-grow-0'
    FlexSmallGrow1 = 'flex-sm-grow-1'
    FlexSmallNowrap = 'flex-sm-nowrap'
    FlexSmallRow = 'flex-sm-row'
    FlexSmallRowReverse = 'flex-sm-row-reverse'
    FlexSmallShrink0 = 'flex-sm-shrink-0'
    FlexSmallShrink1 = 'flex-sm-shrink-1'
    FlexSmallWrap = 'flex-sm-wrap'
    FlexSmallWrapReverse = 'flex-sm-wrap-reverse'
    FlexWrap = 'flex-wrap'
    FlexWrapReverse = 'flex-wrap-reverse'
    FlexExtraLargeColumn = 'flex-xl-column'
    FlexExtraLargeColumnReverse = 'flex-xl-column-reverse'
    FlexExtraLargeFill = 'flex-xl-fill'
    FlexExtraLargeGrow0 = 'flex-xl-grow-0'
    FlexExtraLargeGrow1 = 'flex-xl-grow-1'
    FlexExtraLargeNowrap = 'flex-xl-nowrap'
    FlexExtraLargeRow = 'flex-xl-row'
    FlexExtraLargeRowReverse = 'flex-xl-row-reverse'
    FlexExtraLargeShrink0 = 'flex-xl-shrink-0'
    FlexExtraLargeShrink1 = 'flex-xl-shrink-1'
    FlexExtraLargeWrap = 'flex-xl-wrap'
    FlexExtraLargeWrapReverse = 'flex-xl-wrap-reverse'
    FloatLeft = 'float-left'
    FloatLargeLeft = 'float-lg-left'
    FloatLargeNone = 'float-lg-none'
    FloatLargeRight = 'float-lg-right'
    FloatMediumLeft = 'float-md-left'
    FloatMediumNone = 'float-md-none'
    FloatMediumRight = 'float-md-right'
    FloatNone = 'float-none'
    FloatRight = 'float-right'
    FloatSmallLeft = 'float-sm-left'
    FloatSmallNone = 'float-sm-none'
    FloatSmallRight = 'float-sm-right'
    FloatExtraLargeLeft = 'float-xl-left'
    FloatExtraLargeNone = 'float-xl-none'
    FloatExtraLargeRight = 'float-xl-right'
    Focus = 'focus'
    FontItalic = 'font-italic'
    FontWeightBold = 'font-weight-bold'
    FontWeightBolder = 'font-weight-bolder'
    FontWeightLight = 'font-weight-light'
    FontWeightLighter = 'font-weight-lighter'
    FontWeightNormal = 'font-weight-normal'
    FormCheck = 'form-check'
    FormCheckInline = 'form-check-inline'
    FormCheckInput = 'form-check-input'
    FormCheckLabel = 'form-check-label'
    FormControl = 'form-control'
    FormControlFile = 'form-control-file'
    FormControlLarge = 'form-control-lg'
    FormControlPlaintext = 'form-control-plaintext'
    FormControlRange = 'form-control-range'
    FormControlSmall = 'form-control-sm'
    FormGroup = 'form-group'
    FormInline = 'form-inline'
    FormRow = 'form-row'
    FormText = 'form-text'
    H1 = 'h1'
    Height100 = 'h-100'
    H2 = 'h2'
    Height25 = 'h-25'
    H3 = 'h3'
    H4 = 'h4'
    H5 = 'h5'
    Height50 = 'h-50'
    H6 = 'h6'
    Height75 = 'h-75'
    HeightAuto = 'h-auto'
    Hide = 'hide'
    ImageFluid = 'img-fluid'
    ImageThumbnail = 'img-thumbnail'
    Initialism = 'initialism'
    InputGroup = 'input-group'
    InputGroupAppend = 'input-group-append'
    InputGroupPrepend = 'input-group-prepend'
    InputGroupText = 'input-group-text'
    InvalidFeedback = 'invalid-feedback'
    InvalidTooltip = 'invalid-tooltip'
    Invisible = 'invisible'
    IsInvalid = 'is-invalid'
    IsValid = 'is-valid'
    Jumbotron = 'jumbotron'
    JumbotronFluid = 'jumbotron-fluid'
    JustifyContentAround = 'justify-content-around'
    JustifyContentBetween = 'justify-content-between'
    JustifyContentCenter = 'justify-content-center'
    JustifyContentEnd = 'justify-content-end'
    JustifyContentLargeAround = 'justify-content-lg-around'
    JustifyContentLargeBetween = 'justify-content-lg-between'
    JustifyContentLargeCenter = 'justify-content-lg-center'
    JustifyContentLargeEnd = 'justify-content-lg-end'
    JustifyContentLargeStart = 'justify-content-lg-start'
    JustifyContentMediumAround = 'justify-content-md-around'
    JustifyContentMediumBetween = 'justify-content-md-between'
    JustifyContentMediumCenter = 'justify-content-md-center'
    JustifyContentMediumEnd = 'justify-content-md-end'
    JustifyContentMediumStart = 'justify-content-md-start'
    JustifyContentSmallAround = 'justify-content-sm-around'
    JustifyContentSmallBetween = 'justify-content-sm-between'
    JustifyContentSmallCenter = 'justify-content-sm-center'
    JustifyContentSmallEnd = 'justify-content-sm-end'
    JustifyContentSmallStart = 'justify-content-sm-start'
    JustifyContentStart = 'justify-content-start'
    JustifyContentExtraLargeAround = 'justify-content-xl-around'
    JustifyContentExtraLargeBetween = 'justify-content-xl-between'
    JustifyContentExtraLargeCenter = 'justify-content-xl-center'
    JustifyContentExtraLargeEnd = 'justify-content-xl-end'
    JustifyContentExtraLargeStart = 'justify-content-xl-start'
    Lead = 'lead'
    ListGroup = 'list-group'
    ListGroupHorizontal = 'list-group-horizontal'
    ListGroupHorizontalLarge = 'list-group-horizontal-lg'
    ListGroupHorizontalMedium = 'list-group-horizontal-md'
    ListGroupHorizontalSmall = 'list-group-horizontal-sm'
    ListGroupHorizontalExtraLarge = 'list-group-horizontal-xl'
    ListGroupItem = 'list-group-item'
    ListGroupItemAction = 'list-group-item-action'
    ListGroupItemDanger = 'list-group-item-danger'
    ListGroupItemDark = 'list-group-item-dark'
    ListGroupItemInfo = 'list-group-item-info'
    ListGroupItemLight = 'list-group-item-light'
    ListGroupItemPrimary = 'list-group-item-primary'
    ListGroupItemSecondary = 'list-group-item-secondary'
    ListGroupItemSuccess = 'list-group-item-success'
    ListGroupItemWarning = 'list-group-item-warning'
    ListInline = 'list-inline'
    ListInlineItem = 'list-inline-item'
    ListUnstyled = 'list-unstyled'
    Margin0 = 'm-0'
    Margin1 = 'm-1'
    Margin2 = 'm-2'
    Margin3 = 'm-3'
    Margin4 = 'm-4'
    Margin5 = 'm-5'
    Mark = 'mark'
    MarginAuto = 'm-auto'
    MarginBottom0 = 'mb-0'
    MarginBottom1 = 'mb-1'
    MarginBottom2 = 'mb-2'
    MarginBottom3 = 'mb-3'
    MarginBottom4 = 'mb-4'
    MarginBottom5 = 'mb-5'
    MarginBottomAuto = 'mb-auto'
    MarginBottomLarge0 = 'mb-lg-0'
    MarginBottomLarge1 = 'mb-lg-1'
    MarginBottomLarge2 = 'mb-lg-2'
    MarginBottomLarge3 = 'mb-lg-3'
    MarginBottomLarge4 = 'mb-lg-4'
    MarginBottomLarge5 = 'mb-lg-5'
    MarginBottomLargeAuto = 'mb-lg-auto'
    MarginBottomLargeN1 = 'mb-lg-n1'
    MarginBottomLargeN2 = 'mb-lg-n2'
    MarginBottomLargeN3 = 'mb-lg-n3'
    MarginBottomLargeN4 = 'mb-lg-n4'
    MarginBottomLargeN5 = 'mb-lg-n5'
    MarginBottomMedium0 = 'mb-md-0'
    MarginBottomMedium1 = 'mb-md-1'
    MarginBottomMedium2 = 'mb-md-2'
    MarginBottomMedium3 = 'mb-md-3'
    MarginBottomMedium4 = 'mb-md-4'
    MarginBottomMedium5 = 'mb-md-5'
    MarginBottomMediumAuto = 'mb-md-auto'
    MarginBottomMediumN1 = 'mb-md-n1'
    MarginBottomMediumN2 = 'mb-md-n2'
    MarginBottomMediumN3 = 'mb-md-n3'
    MarginBottomMediumN4 = 'mb-md-n4'
    MarginBottomMediumN5 = 'mb-md-n5'
    MarginBottomN1 = 'mb-n1'
    MarginBottomN2 = 'mb-n2'
    MarginBottomN3 = 'mb-n3'
    MarginBottomN4 = 'mb-n4'
    MarginBottomN5 = 'mb-n5'
    MarginBottomSmall0 = 'mb-sm-0'
    MarginBottomSmall1 = 'mb-sm-1'
    MarginBottomSmall2 = 'mb-sm-2'
    MarginBottomSmall3 = 'mb-sm-3'
    MarginBottomSmall4 = 'mb-sm-4'
    MarginBottomSmall5 = 'mb-sm-5'
    MarginBottomSmallAuto = 'mb-sm-auto'
    MarginBottomSmallN1 = 'mb-sm-n1'
    MarginBottomSmallN2 = 'mb-sm-n2'
    MarginBottomSmallN3 = 'mb-sm-n3'
    MarginBottomSmallN4 = 'mb-sm-n4'
    MarginBottomSmallN5 = 'mb-sm-n5'
    MarginBottomExtraLarge0 = 'mb-xl-0'
    MarginBottomExtraLarge1 = 'mb-xl-1'
    MarginBottomExtraLarge2 = 'mb-xl-2'
    MarginBottomExtraLarge3 = 'mb-xl-3'
    MarginBottomExtraLarge4 = 'mb-xl-4'
    MarginBottomExtraLarge5 = 'mb-xl-5'
    MarginBottomExtraLargeAuto = 'mb-xl-auto'
    MarginBottomExtraLargeN1 = 'mb-xl-n1'
    MarginBottomExtraLargeN2 = 'mb-xl-n2'
    MarginBottomExtraLargeN3 = 'mb-xl-n3'
    MarginBottomExtraLargeN4 = 'mb-xl-n4'
    MarginBottomExtraLargeN5 = 'mb-xl-n5'
    Media = 'media'
    MediaBody = 'media-body'
    Mh100 = 'mh-100'
    MinVh100 = 'min-vh-100'
    MinVw100 = 'min-vw-100'
    MarginLeft0 = 'ml-0'
    MarginLeft1 = 'ml-1'
    MarginLeft2 = 'ml-2'
    MarginLeft3 = 'ml-3'
    MarginLeft4 = 'ml-4'
    MarginLeft5 = 'ml-5'
    MarginLeftAuto = 'ml-auto'
    MarginLarge0 = 'm-lg-0'
    MarginLarge1 = 'm-lg-1'
    MarginLarge2 = 'm-lg-2'
    MarginLarge3 = 'm-lg-3'
    MarginLarge4 = 'm-lg-4'
    MarginLarge5 = 'm-lg-5'
    MarginLargeAuto = 'm-lg-auto'
    MarginLargeN1 = 'm-lg-n1'
    MarginLargeN2 = 'm-lg-n2'
    MarginLargeN3 = 'm-lg-n3'
    MarginLargeN4 = 'm-lg-n4'
    MarginLargeN5 = 'm-lg-n5'
    MarginLeftLarge0 = 'ml-lg-0'
    MarginLeftLarge1 = 'ml-lg-1'
    MarginLeftLarge2 = 'ml-lg-2'
    MarginLeftLarge3 = 'ml-lg-3'
    MarginLeftLarge4 = 'ml-lg-4'
    MarginLeftLarge5 = 'ml-lg-5'
    MarginLeftLargeAuto = 'ml-lg-auto'
    MarginLeftLargeN1 = 'ml-lg-n1'
    MarginLeftLargeN2 = 'ml-lg-n2'
    MarginLeftLargeN3 = 'ml-lg-n3'
    MarginLeftLargeN4 = 'ml-lg-n4'
    MarginLeftLargeN5 = 'ml-lg-n5'
    MarginLeftMedium0 = 'ml-md-0'
    MarginLeftMedium1 = 'ml-md-1'
    MarginLeftMedium2 = 'ml-md-2'
    MarginLeftMedium3 = 'ml-md-3'
    MarginLeftMedium4 = 'ml-md-4'
    MarginLeftMedium5 = 'ml-md-5'
    MarginLeftMediumAuto = 'ml-md-auto'
    MarginLeftMediumN1 = 'ml-md-n1'
    MarginLeftMediumN2 = 'ml-md-n2'
    MarginLeftMediumN3 = 'ml-md-n3'
    MarginLeftMediumN4 = 'ml-md-n4'
    MarginLeftMediumN5 = 'ml-md-n5'
    MarginLeftN1 = 'ml-n1'
    MarginLeftN2 = 'ml-n2'
    MarginLeftN3 = 'ml-n3'
    MarginLeftN4 = 'ml-n4'
    MarginLeftN5 = 'ml-n5'
    MarginLeftSmall0 = 'ml-sm-0'
    MarginLeftSmall1 = 'ml-sm-1'
    MarginLeftSmall2 = 'ml-sm-2'
    MarginLeftSmall3 = 'ml-sm-3'
    MarginLeftSmall4 = 'ml-sm-4'
    MarginLeftSmall5 = 'ml-sm-5'
    MarginLeftSmallAuto = 'ml-sm-auto'
    MarginLeftSmallN1 = 'ml-sm-n1'
    MarginLeftSmallN2 = 'ml-sm-n2'
    MarginLeftSmallN3 = 'ml-sm-n3'
    MarginLeftSmallN4 = 'ml-sm-n4'
    MarginLeftSmallN5 = 'ml-sm-n5'
    MarginLeftExtraLarge0 = 'ml-xl-0'
    MarginLeftExtraLarge1 = 'ml-xl-1'
    MarginLeftExtraLarge2 = 'ml-xl-2'
    MarginLeftExtraLarge3 = 'ml-xl-3'
    MarginLeftExtraLarge4 = 'ml-xl-4'
    MarginLeftExtraLarge5 = 'ml-xl-5'
    MarginLeftExtraLargeAuto = 'ml-xl-auto'
    MarginLeftExtraLargeN1 = 'ml-xl-n1'
    MarginLeftExtraLargeN2 = 'ml-xl-n2'
    MarginLeftExtraLargeN3 = 'ml-xl-n3'
    MarginLeftExtraLargeN4 = 'ml-xl-n4'
    MarginLeftExtraLargeN5 = 'ml-xl-n5'
    MarginMedium0 = 'm-md-0'
    MarginMedium1 = 'm-md-1'
    MarginMedium2 = 'm-md-2'
    MarginMedium3 = 'm-md-3'
    MarginMedium4 = 'm-md-4'
    MarginMedium5 = 'm-md-5'
    MarginMediumAuto = 'm-md-auto'
    MarginMediumN1 = 'm-md-n1'
    MarginMediumN2 = 'm-md-n2'
    MarginMediumN3 = 'm-md-n3'
    MarginMediumN4 = 'm-md-n4'
    MarginMediumN5 = 'm-md-n5'
    MarginN1 = 'm-n1'
    MarginN2 = 'm-n2'
    MarginN3 = 'm-n3'
    MarginN4 = 'm-n4'
    MarginN5 = 'm-n5'
    Modal = 'modal'
    ModalBackdrop = 'modal-backdrop'
    ModalBody = 'modal-body'
    ModalContent = 'modal-content'
    ModalDialog = 'modal-dialog'
    ModalDialogCentered = 'modal-dialog-centered'
    ModalDialogScrollable = 'modal-dialog-scrollable'
    ModalFooter = 'modal-footer'
    ModalHeader = 'modal-header'
    ModalLarge = 'modal-lg'
    ModalOpen = 'modal-open'
    ModalScrollbarMeasure = 'modal-scrollbar-measure'
    ModalSmall = 'modal-sm'
    ModalTitle = 'modal-title'
    ModalExtraLarge = 'modal-xl'
    MarginRight0 = 'mr-0'
    MarginRight1 = 'mr-1'
    MarginRight2 = 'mr-2'
    MarginRight3 = 'mr-3'
    MarginRight4 = 'mr-4'
    MarginRight5 = 'mr-5'
    MarginRightAuto = 'mr-auto'
    MarginRightLarge0 = 'mr-lg-0'
    MarginRightLarge1 = 'mr-lg-1'
    MarginRightLarge2 = 'mr-lg-2'
    MarginRightLarge3 = 'mr-lg-3'
    MarginRightLarge4 = 'mr-lg-4'
    MarginRightLarge5 = 'mr-lg-5'
    MarginRightLargeAuto = 'mr-lg-auto'
    MarginRightLargeN1 = 'mr-lg-n1'
    MarginRightLargeN2 = 'mr-lg-n2'
    MarginRightLargeN3 = 'mr-lg-n3'
    MarginRightLargeN4 = 'mr-lg-n4'
    MarginRightLargeN5 = 'mr-lg-n5'
    MarginRightMedium0 = 'mr-md-0'
    MarginRightMedium1 = 'mr-md-1'
    MarginRightMedium2 = 'mr-md-2'
    MarginRightMedium3 = 'mr-md-3'
    MarginRightMedium4 = 'mr-md-4'
    MarginRightMedium5 = 'mr-md-5'
    MarginRightMediumAuto = 'mr-md-auto'
    MarginRightMediumN1 = 'mr-md-n1'
    MarginRightMediumN2 = 'mr-md-n2'
    MarginRightMediumN3 = 'mr-md-n3'
    MarginRightMediumN4 = 'mr-md-n4'
    MarginRightMediumN5 = 'mr-md-n5'
    MarginRightN1 = 'mr-n1'
    MarginRightN2 = 'mr-n2'
    MarginRightN3 = 'mr-n3'
    MarginRightN4 = 'mr-n4'
    MarginRightN5 = 'mr-n5'
    MarginRightSmall0 = 'mr-sm-0'
    MarginRightSmall1 = 'mr-sm-1'
    MarginRightSmall2 = 'mr-sm-2'
    MarginRightSmall3 = 'mr-sm-3'
    MarginRightSmall4 = 'mr-sm-4'
    MarginRightSmall5 = 'mr-sm-5'
    MarginRightSmallAuto = 'mr-sm-auto'
    MarginRightSmallN1 = 'mr-sm-n1'
    MarginRightSmallN2 = 'mr-sm-n2'
    MarginRightSmallN3 = 'mr-sm-n3'
    MarginRightSmallN4 = 'mr-sm-n4'
    MarginRightSmallN5 = 'mr-sm-n5'
    MarginRightExtraLarge0 = 'mr-xl-0'
    MarginRightExtraLarge1 = 'mr-xl-1'
    MarginRightExtraLarge2 = 'mr-xl-2'
    MarginRightExtraLarge3 = 'mr-xl-3'
    MarginRightExtraLarge4 = 'mr-xl-4'
    MarginRightExtraLarge5 = 'mr-xl-5'
    MarginRightExtraLargeAuto = 'mr-xl-auto'
    MarginRightExtraLargeN1 = 'mr-xl-n1'
    MarginRightExtraLargeN2 = 'mr-xl-n2'
    MarginRightExtraLargeN3 = 'mr-xl-n3'
    MarginRightExtraLargeN4 = 'mr-xl-n4'
    MarginRightExtraLargeN5 = 'mr-xl-n5'
    MarginSmall0 = 'm-sm-0'
    MarginSmall1 = 'm-sm-1'
    MarginSmall2 = 'm-sm-2'
    MarginSmall3 = 'm-sm-3'
    MarginSmall4 = 'm-sm-4'
    MarginSmall5 = 'm-sm-5'
    MarginSmallAuto = 'm-sm-auto'
    MarginSmallN1 = 'm-sm-n1'
    MarginSmallN2 = 'm-sm-n2'
    MarginSmallN3 = 'm-sm-n3'
    MarginSmallN4 = 'm-sm-n4'
    MarginSmallN5 = 'm-sm-n5'
    MarginTop0 = 'mt-0'
    MarginTop1 = 'mt-1'
    MarginTop2 = 'mt-2'
    MarginTop3 = 'mt-3'
    MarginTop4 = 'mt-4'
    MarginTop5 = 'mt-5'
    MarginTopAuto = 'mt-auto'
    MarginTopLarge0 = 'mt-lg-0'
    MarginTopLarge1 = 'mt-lg-1'
    MarginTopLarge2 = 'mt-lg-2'
    MarginTopLarge3 = 'mt-lg-3'
    MarginTopLarge4 = 'mt-lg-4'
    MarginTopLarge5 = 'mt-lg-5'
    MarginTopLargeAuto = 'mt-lg-auto'
    MarginTopLargeN1 = 'mt-lg-n1'
    MarginTopLargeN2 = 'mt-lg-n2'
    MarginTopLargeN3 = 'mt-lg-n3'
    MarginTopLargeN4 = 'mt-lg-n4'
    MarginTopLargeN5 = 'mt-lg-n5'
    MarginTopMedium0 = 'mt-md-0'
    MarginTopMedium1 = 'mt-md-1'
    MarginTopMedium2 = 'mt-md-2'
    MarginTopMedium3 = 'mt-md-3'
    MarginTopMedium4 = 'mt-md-4'
    MarginTopMedium5 = 'mt-md-5'
    MarginTopMediumAuto = 'mt-md-auto'
    MarginTopMediumN1 = 'mt-md-n1'
    MarginTopMediumN2 = 'mt-md-n2'
    MarginTopMediumN3 = 'mt-md-n3'
    MarginTopMediumN4 = 'mt-md-n4'
    MarginTopMediumN5 = 'mt-md-n5'
    MarginTopN1 = 'mt-n1'
    MarginTopN2 = 'mt-n2'
    MarginTopN3 = 'mt-n3'
    MarginTopN4 = 'mt-n4'
    MarginTopN5 = 'mt-n5'
    MarginTopSmall0 = 'mt-sm-0'
    MarginTopSmall1 = 'mt-sm-1'
    MarginTopSmall2 = 'mt-sm-2'
    MarginTopSmall3 = 'mt-sm-3'
    MarginTopSmall4 = 'mt-sm-4'
    MarginTopSmall5 = 'mt-sm-5'
    MarginTopSmallAuto = 'mt-sm-auto'
    MarginTopSmallN1 = 'mt-sm-n1'
    MarginTopSmallN2 = 'mt-sm-n2'
    MarginTopSmallN3 = 'mt-sm-n3'
    MarginTopSmallN4 = 'mt-sm-n4'
    MarginTopSmallN5 = 'mt-sm-n5'
    MarginTopExtraLarge0 = 'mt-xl-0'
    MarginTopExtraLarge1 = 'mt-xl-1'
    MarginTopExtraLarge2 = 'mt-xl-2'
    MarginTopExtraLarge3 = 'mt-xl-3'
    MarginTopExtraLarge4 = 'mt-xl-4'
    MarginTopExtraLarge5 = 'mt-xl-5'
    MarginTopExtraLargeAuto = 'mt-xl-auto'
    MarginTopExtraLargeN1 = 'mt-xl-n1'
    MarginTopExtraLargeN2 = 'mt-xl-n2'
    MarginTopExtraLargeN3 = 'mt-xl-n3'
    MarginTopExtraLargeN4 = 'mt-xl-n4'
    MarginTopExtraLargeN5 = 'mt-xl-n5'
    Mw100 = 'mw-100'
    MarginX0 = 'mx-0'
    MarginX1 = 'mx-1'
    MarginX2 = 'mx-2'
    MarginX3 = 'mx-3'
    MarginX4 = 'mx-4'
    MarginX5 = 'mx-5'
    MarginXAuto = 'mx-auto'
    MarginExtraLarge0 = 'm-xl-0'
    MarginExtraLarge1 = 'm-xl-1'
    MarginExtraLarge2 = 'm-xl-2'
    MarginExtraLarge3 = 'm-xl-3'
    MarginExtraLarge4 = 'm-xl-4'
    MarginExtraLarge5 = 'm-xl-5'
    MarginExtraLargeAuto = 'm-xl-auto'
    MarginXLarge0 = 'mx-lg-0'
    MarginXLarge1 = 'mx-lg-1'
    MarginXLarge2 = 'mx-lg-2'
    MarginXLarge3 = 'mx-lg-3'
    MarginXLarge4 = 'mx-lg-4'
    MarginXLarge5 = 'mx-lg-5'
    MarginXLargeAuto = 'mx-lg-auto'
    MarginXLargeN1 = 'mx-lg-n1'
    MarginXLargeN2 = 'mx-lg-n2'
    MarginXLargeN3 = 'mx-lg-n3'
    MarginXLargeN4 = 'mx-lg-n4'
    MarginXLargeN5 = 'mx-lg-n5'
    MarginExtraLargeN1 = 'm-xl-n1'
    MarginExtraLargeN2 = 'm-xl-n2'
    MarginExtraLargeN3 = 'm-xl-n3'
    MarginExtraLargeN4 = 'm-xl-n4'
    MarginExtraLargeN5 = 'm-xl-n5'
    MarginXMedium0 = 'mx-md-0'
    MarginXMedium1 = 'mx-md-1'
    MarginXMedium2 = 'mx-md-2'
    MarginXMedium3 = 'mx-md-3'
    MarginXMedium4 = 'mx-md-4'
    MarginXMedium5 = 'mx-md-5'
    MarginXMediumAuto = 'mx-md-auto'
    MarginXMediumN1 = 'mx-md-n1'
    MarginXMediumN2 = 'mx-md-n2'
    MarginXMediumN3 = 'mx-md-n3'
    MarginXMediumN4 = 'mx-md-n4'
    MarginXMediumN5 = 'mx-md-n5'
    MarginXN1 = 'mx-n1'
    MarginXN2 = 'mx-n2'
    MarginXN3 = 'mx-n3'
    MarginXN4 = 'mx-n4'
    MarginXN5 = 'mx-n5'
    MarginXSmall0 = 'mx-sm-0'
    MarginXSmall1 = 'mx-sm-1'
    MarginXSmall2 = 'mx-sm-2'
    MarginXSmall3 = 'mx-sm-3'
    MarginXSmall4 = 'mx-sm-4'
    MarginXSmall5 = 'mx-sm-5'
    MarginXSmallAuto = 'mx-sm-auto'
    MarginXSmallN1 = 'mx-sm-n1'
    MarginXSmallN2 = 'mx-sm-n2'
    MarginXSmallN3 = 'mx-sm-n3'
    MarginXSmallN4 = 'mx-sm-n4'
    MarginXSmallN5 = 'mx-sm-n5'
    MarginXExtraLarge0 = 'mx-xl-0'
    MarginXExtraLarge1 = 'mx-xl-1'
    MarginXExtraLarge2 = 'mx-xl-2'
    MarginXExtraLarge3 = 'mx-xl-3'
    MarginXExtraLarge4 = 'mx-xl-4'
    MarginXExtraLarge5 = 'mx-xl-5'
    MarginXExtraLargeAuto = 'mx-xl-auto'
    MarginXExtraLargeN1 = 'mx-xl-n1'
    MarginXExtraLargeN2 = 'mx-xl-n2'
    MarginXExtraLargeN3 = 'mx-xl-n3'
    MarginXExtraLargeN4 = 'mx-xl-n4'
    MarginXExtraLargeN5 = 'mx-xl-n5'
    MarginY0 = 'my-0'
    MarginY1 = 'my-1'
    MarginY2 = 'my-2'
    MarginY3 = 'my-3'
    MarginY4 = 'my-4'
    MarginY5 = 'my-5'
    MarginYAuto = 'my-auto'
    MarginYLarge0 = 'my-lg-0'
    MarginYLarge1 = 'my-lg-1'
    MarginYLarge2 = 'my-lg-2'
    MarginYLarge3 = 'my-lg-3'
    MarginYLarge4 = 'my-lg-4'
    MarginYLarge5 = 'my-lg-5'
    MarginYLargeAuto = 'my-lg-auto'
    MarginYLargeN1 = 'my-lg-n1'
    MarginYLargeN2 = 'my-lg-n2'
    MarginYLargeN3 = 'my-lg-n3'
    MarginYLargeN4 = 'my-lg-n4'
    MarginYLargeN5 = 'my-lg-n5'
    MarginYMedium0 = 'my-md-0'
    MarginYMedium1 = 'my-md-1'
    MarginYMedium2 = 'my-md-2'
    MarginYMedium3 = 'my-md-3'
    MarginYMedium4 = 'my-md-4'
    MarginYMedium5 = 'my-md-5'
    MarginYMediumAuto = 'my-md-auto'
    MarginYMediumN1 = 'my-md-n1'
    MarginYMediumN2 = 'my-md-n2'
    MarginYMediumN3 = 'my-md-n3'
    MarginYMediumN4 = 'my-md-n4'
    MarginYMediumN5 = 'my-md-n5'
    MarginYN1 = 'my-n1'
    MarginYN2 = 'my-n2'
    MarginYN3 = 'my-n3'
    MarginYN4 = 'my-n4'
    MarginYN5 = 'my-n5'
    MarginYSmall0 = 'my-sm-0'
    MarginYSmall1 = 'my-sm-1'
    MarginYSmall2 = 'my-sm-2'
    MarginYSmall3 = 'my-sm-3'
    MarginYSmall4 = 'my-sm-4'
    MarginYSmall5 = 'my-sm-5'
    MarginYSmallAuto = 'my-sm-auto'
    MarginYSmallN1 = 'my-sm-n1'
    MarginYSmallN2 = 'my-sm-n2'
    MarginYSmallN3 = 'my-sm-n3'
    MarginYSmallN4 = 'my-sm-n4'
    MarginYSmallN5 = 'my-sm-n5'
    MarginYExtraLarge0 = 'my-xl-0'
    MarginYExtraLarge1 = 'my-xl-1'
    MarginYExtraLarge2 = 'my-xl-2'
    MarginYExtraLarge3 = 'my-xl-3'
    MarginYExtraLarge4 = 'my-xl-4'
    MarginYExtraLarge5 = 'my-xl-5'
    MarginYExtraLargeAuto = 'my-xl-auto'
    MarginYExtraLargeN1 = 'my-xl-n1'
    MarginYExtraLargeN2 = 'my-xl-n2'
    MarginYExtraLargeN3 = 'my-xl-n3'
    MarginYExtraLargeN4 = 'my-xl-n4'
    MarginYExtraLargeN5 = 'my-xl-n5'
    Nav = 'nav'
    Navbar = 'navbar'
    NavbarBrand = 'navbar-brand'
    NavbarCollapse = 'navbar-collapse'
    NavbarExpand = 'navbar-expand'
    NavbarExpandLarge = 'navbar-expand-lg'
    NavbarExpandMedium = 'navbar-expand-md'
    NavbarExpandSmall = 'navbar-expand-sm'
    NavbarExpandExtraLarge = 'navbar-expand-xl'
    NavbarNav = 'navbar-nav'
    NavbarText = 'navbar-text'
    NavbarToggler = 'navbar-toggler'
    NavbarTogglerIcon = 'navbar-toggler-icon'
    NavItem = 'nav-item'
    NavLink = 'nav-link'
    NavTabs = 'nav-tabs'
    NoGutters = 'no-gutters'
    Offset1 = 'offset-1'
    Offset10 = 'offset-10'
    Offset11 = 'offset-11'
    Offset2 = 'offset-2'
    Offset3 = 'offset-3'
    Offset4 = 'offset-4'
    Offset5 = 'offset-5'
    Offset6 = 'offset-6'
    Offset7 = 'offset-7'
    Offset8 = 'offset-8'
    Offset9 = 'offset-9'
    OffsetLarge0 = 'offset-lg-0'
    OffsetLarge1 = 'offset-lg-1'
    OffsetLarge10 = 'offset-lg-10'
    OffsetLarge11 = 'offset-lg-11'
    OffsetLarge2 = 'offset-lg-2'
    OffsetLarge3 = 'offset-lg-3'
    OffsetLarge4 = 'offset-lg-4'
    OffsetLarge5 = 'offset-lg-5'
    OffsetLarge6 = 'offset-lg-6'
    OffsetLarge7 = 'offset-lg-7'
    OffsetLarge8 = 'offset-lg-8'
    OffsetLarge9 = 'offset-lg-9'
    OffsetMedium0 = 'offset-md-0'
    OffsetMedium1 = 'offset-md-1'
    OffsetMedium10 = 'offset-md-10'
    OffsetMedium11 = 'offset-md-11'
    OffsetMedium2 = 'offset-md-2'
    OffsetMedium3 = 'offset-md-3'
    OffsetMedium4 = 'offset-md-4'
    OffsetMedium5 = 'offset-md-5'
    OffsetMedium6 = 'offset-md-6'
    OffsetMedium7 = 'offset-md-7'
    OffsetMedium8 = 'offset-md-8'
    OffsetMedium9 = 'offset-md-9'
    OffsetSmall0 = 'offset-sm-0'
    OffsetSmall1 = 'offset-sm-1'
    OffsetSmall10 = 'offset-sm-10'
    OffsetSmall11 = 'offset-sm-11'
    OffsetSmall2 = 'offset-sm-2'
    OffsetSmall3 = 'offset-sm-3'
    OffsetSmall4 = 'offset-sm-4'
    OffsetSmall5 = 'offset-sm-5'
    OffsetSmall6 = 'offset-sm-6'
    OffsetSmall7 = 'offset-sm-7'
    OffsetSmall8 = 'offset-sm-8'
    OffsetSmall9 = 'offset-sm-9'
    OffsetExtraLarge0 = 'offset-xl-0'
    OffsetExtraLarge1 = 'offset-xl-1'
    OffsetExtraLarge10 = 'offset-xl-10'
    OffsetExtraLarge11 = 'offset-xl-11'
    OffsetExtraLarge2 = 'offset-xl-2'
    OffsetExtraLarge3 = 'offset-xl-3'
    OffsetExtraLarge4 = 'offset-xl-4'
    OffsetExtraLarge5 = 'offset-xl-5'
    OffsetExtraLarge6 = 'offset-xl-6'
    OffsetExtraLarge7 = 'offset-xl-7'
    OffsetExtraLarge8 = 'offset-xl-8'
    OffsetExtraLarge9 = 'offset-xl-9'
    Order0 = 'order-0'
    Order1 = 'order-1'
    Order10 = 'order-10'
    Order11 = 'order-11'
    Order12 = 'order-12'
    Order2 = 'order-2'
    Order3 = 'order-3'
    Order4 = 'order-4'
    Order5 = 'order-5'
    Order6 = 'order-6'
    Order7 = 'order-7'
    Order8 = 'order-8'
    Order9 = 'order-9'
    OrderFirst = 'order-first'
    OrderLast = 'order-last'
    OrderLarge0 = 'order-lg-0'
    OrderLarge1 = 'order-lg-1'
    OrderLarge10 = 'order-lg-10'
    OrderLarge11 = 'order-lg-11'
    OrderLarge12 = 'order-lg-12'
    OrderLarge2 = 'order-lg-2'
    OrderLarge3 = 'order-lg-3'
    OrderLarge4 = 'order-lg-4'
    OrderLarge5 = 'order-lg-5'
    OrderLarge6 = 'order-lg-6'
    OrderLarge7 = 'order-lg-7'
    OrderLarge8 = 'order-lg-8'
    OrderLarge9 = 'order-lg-9'
    OrderLargeFirst = 'order-lg-first'
    OrderLargeLast = 'order-lg-last'
    OrderMedium0 = 'order-md-0'
    OrderMedium1 = 'order-md-1'
    OrderMedium10 = 'order-md-10'
    OrderMedium11 = 'order-md-11'
    OrderMedium12 = 'order-md-12'
    OrderMedium2 = 'order-md-2'
    OrderMedium3 = 'order-md-3'
    OrderMedium4 = 'order-md-4'
    OrderMedium5 = 'order-md-5'
    OrderMedium6 = 'order-md-6'
    OrderMedium7 = 'order-md-7'
    OrderMedium8 = 'order-md-8'
    OrderMedium9 = 'order-md-9'
    OrderMediumFirst = 'order-md-first'
    OrderMediumLast = 'order-md-last'
    OrderSmall0 = 'order-sm-0'
    OrderSmall1 = 'order-sm-1'
    OrderSmall10 = 'order-sm-10'
    OrderSmall11 = 'order-sm-11'
    OrderSmall12 = 'order-sm-12'
    OrderSmall2 = 'order-sm-2'
    OrderSmall3 = 'order-sm-3'
    OrderSmall4 = 'order-sm-4'
    OrderSmall5 = 'order-sm-5'
    OrderSmall6 = 'order-sm-6'
    OrderSmall7 = 'order-sm-7'
    OrderSmall8 = 'order-sm-8'
    OrderSmall9 = 'order-sm-9'
    OrderSmallFirst = 'order-sm-first'
    OrderSmallLast = 'order-sm-last'
    OrderExtraLarge0 = 'order-xl-0'
    OrderExtraLarge1 = 'order-xl-1'
    OrderExtraLarge10 = 'order-xl-10'
    OrderExtraLarge11 = 'order-xl-11'
    OrderExtraLarge12 = 'order-xl-12'
    OrderExtraLarge2 = 'order-xl-2'
    OrderExtraLarge3 = 'order-xl-3'
    OrderExtraLarge4 = 'order-xl-4'
    OrderExtraLarge5 = 'order-xl-5'
    OrderExtraLarge6 = 'order-xl-6'
    OrderExtraLarge7 = 'order-xl-7'
    OrderExtraLarge8 = 'order-xl-8'
    OrderExtraLarge9 = 'order-xl-9'
    OrderExtraLargeFirst = 'order-xl-first'
    OrderExtraLargeLast = 'order-xl-last'
    OverflowAuto = 'overflow-auto'
    OverflowHidden = 'overflow-hidden'
    Padding0 = 'p-0'
    Padding1 = 'p-1'
    Padding2 = 'p-2'
    Padding3 = 'p-3'
    Padding4 = 'p-4'
    Padding5 = 'p-5'
    PageLink = 'page-link'
    Pagination = 'pagination'
    PaddingBottom0 = 'pb-0'
    PaddingBottom1 = 'pb-1'
    PaddingBottom2 = 'pb-2'
    PaddingBottom3 = 'pb-3'
    PaddingBottom4 = 'pb-4'
    PaddingBottom5 = 'pb-5'
    PaddingBottomLarge0 = 'pb-lg-0'
    PaddingBottomLarge1 = 'pb-lg-1'
    PaddingBottomLarge2 = 'pb-lg-2'
    PaddingBottomLarge3 = 'pb-lg-3'
    PaddingBottomLarge4 = 'pb-lg-4'
    PaddingBottomLarge5 = 'pb-lg-5'
    PaddingBottomMedium0 = 'pb-md-0'
    PaddingBottomMedium1 = 'pb-md-1'
    PaddingBottomMedium2 = 'pb-md-2'
    PaddingBottomMedium3 = 'pb-md-3'
    PaddingBottomMedium4 = 'pb-md-4'
    PaddingBottomMedium5 = 'pb-md-5'
    PaddingBottomSmall0 = 'pb-sm-0'
    PaddingBottomSmall1 = 'pb-sm-1'
    PaddingBottomSmall2 = 'pb-sm-2'
    PaddingBottomSmall3 = 'pb-sm-3'
    PaddingBottomSmall4 = 'pb-sm-4'
    PaddingBottomSmall5 = 'pb-sm-5'
    PaddingBottomExtraLarge0 = 'pb-xl-0'
    PaddingBottomExtraLarge1 = 'pb-xl-1'
    PaddingBottomExtraLarge2 = 'pb-xl-2'
    PaddingBottomExtraLarge3 = 'pb-xl-3'
    PaddingBottomExtraLarge4 = 'pb-xl-4'
    PaddingBottomExtraLarge5 = 'pb-xl-5'
    PaddingLeft0 = 'pl-0'
    PaddingLeft1 = 'pl-1'
    PaddingLeft2 = 'pl-2'
    PaddingLeft3 = 'pl-3'
    PaddingLeft4 = 'pl-4'
    PaddingLeft5 = 'pl-5'
    PaddingLarge0 = 'p-lg-0'
    PaddingLarge1 = 'p-lg-1'
    PaddingLarge2 = 'p-lg-2'
    PaddingLarge3 = 'p-lg-3'
    PaddingLarge4 = 'p-lg-4'
    PaddingLarge5 = 'p-lg-5'
    PaddingLeftLarge0 = 'pl-lg-0'
    PaddingLeftLarge1 = 'pl-lg-1'
    PaddingLeftLarge2 = 'pl-lg-2'
    PaddingLeftLarge3 = 'pl-lg-3'
    PaddingLeftLarge4 = 'pl-lg-4'
    PaddingLeftLarge5 = 'pl-lg-5'
    PaddingLeftMedium0 = 'pl-md-0'
    PaddingLeftMedium1 = 'pl-md-1'
    PaddingLeftMedium2 = 'pl-md-2'
    PaddingLeftMedium3 = 'pl-md-3'
    PaddingLeftMedium4 = 'pl-md-4'
    PaddingLeftMedium5 = 'pl-md-5'
    PaddingLeftSmall0 = 'pl-sm-0'
    PaddingLeftSmall1 = 'pl-sm-1'
    PaddingLeftSmall2 = 'pl-sm-2'
    PaddingLeftSmall3 = 'pl-sm-3'
    PaddingLeftSmall4 = 'pl-sm-4'
    PaddingLeftSmall5 = 'pl-sm-5'
    PaddingLeftExtraLarge0 = 'pl-xl-0'
    PaddingLeftExtraLarge1 = 'pl-xl-1'
    PaddingLeftExtraLarge2 = 'pl-xl-2'
    PaddingLeftExtraLarge3 = 'pl-xl-3'
    PaddingLeftExtraLarge4 = 'pl-xl-4'
    PaddingLeftExtraLarge5 = 'pl-xl-5'
    PaddingMedium0 = 'p-md-0'
    PaddingMedium1 = 'p-md-1'
    PaddingMedium2 = 'p-md-2'
    PaddingMedium3 = 'p-md-3'
    PaddingMedium4 = 'p-md-4'
    PaddingMedium5 = 'p-md-5'
    PointerEvent = 'pointer-event'
    Popover = 'popover'
    PopoverBody = 'popover-body'
    PopoverHeader = 'popover-header'
    PositionAbsolute = 'position-absolute'
    PositionFixed = 'position-fixed'
    PositionRelative = 'position-relative'
    PositionStatic = 'position-static'
    PositionSticky = 'position-sticky'
    PaddingRight0 = 'pr-0'
    PaddingRight1 = 'pr-1'
    PaddingRight2 = 'pr-2'
    PaddingRight3 = 'pr-3'
    PaddingRight4 = 'pr-4'
    PaddingRight5 = 'pr-5'
    PreScrollable = 'pre-scrollable'
    PaddingRightLarge0 = 'pr-lg-0'
    PaddingRightLarge1 = 'pr-lg-1'
    PaddingRightLarge2 = 'pr-lg-2'
    PaddingRightLarge3 = 'pr-lg-3'
    PaddingRightLarge4 = 'pr-lg-4'
    PaddingRightLarge5 = 'pr-lg-5'
    PaddingRightMedium0 = 'pr-md-0'
    PaddingRightMedium1 = 'pr-md-1'
    PaddingRightMedium2 = 'pr-md-2'
    PaddingRightMedium3 = 'pr-md-3'
    PaddingRightMedium4 = 'pr-md-4'
    PaddingRightMedium5 = 'pr-md-5'
    Progress = 'progress'
    ProgressBar = 'progress-bar'
    ProgressBarAnimated = 'progress-bar-animated'
    ProgressBarStriped = 'progress-bar-striped'
    PaddingRightSmall0 = 'pr-sm-0'
    PaddingRightSmall1 = 'pr-sm-1'
    PaddingRightSmall2 = 'pr-sm-2'
    PaddingRightSmall3 = 'pr-sm-3'
    PaddingRightSmall4 = 'pr-sm-4'
    PaddingRightSmall5 = 'pr-sm-5'
    PaddingRightExtraLarge0 = 'pr-xl-0'
    PaddingRightExtraLarge1 = 'pr-xl-1'
    PaddingRightExtraLarge2 = 'pr-xl-2'
    PaddingRightExtraLarge3 = 'pr-xl-3'
    PaddingRightExtraLarge4 = 'pr-xl-4'
    PaddingRightExtraLarge5 = 'pr-xl-5'
    PaddingSmall0 = 'p-sm-0'
    PaddingSmall1 = 'p-sm-1'
    PaddingSmall2 = 'p-sm-2'
    PaddingSmall3 = 'p-sm-3'
    PaddingSmall4 = 'p-sm-4'
    PaddingSmall5 = 'p-sm-5'
    PaddingTop0 = 'pt-0'
    PaddingTop1 = 'pt-1'
    PaddingTop2 = 'pt-2'
    PaddingTop3 = 'pt-3'
    PaddingTop4 = 'pt-4'
    PaddingTop5 = 'pt-5'
    PaddingTopLarge0 = 'pt-lg-0'
    PaddingTopLarge1 = 'pt-lg-1'
    PaddingTopLarge2 = 'pt-lg-2'
    PaddingTopLarge3 = 'pt-lg-3'
    PaddingTopLarge4 = 'pt-lg-4'
    PaddingTopLarge5 = 'pt-lg-5'
    PaddingTopMedium0 = 'pt-md-0'
    PaddingTopMedium1 = 'pt-md-1'
    PaddingTopMedium2 = 'pt-md-2'
    PaddingTopMedium3 = 'pt-md-3'
    PaddingTopMedium4 = 'pt-md-4'
    PaddingTopMedium5 = 'pt-md-5'
    PaddingTopSmall0 = 'pt-sm-0'
    PaddingTopSmall1 = 'pt-sm-1'
    PaddingTopSmall2 = 'pt-sm-2'
    PaddingTopSmall3 = 'pt-sm-3'
    PaddingTopSmall4 = 'pt-sm-4'
    PaddingTopSmall5 = 'pt-sm-5'
    PaddingTopExtraLarge0 = 'pt-xl-0'
    PaddingTopExtraLarge1 = 'pt-xl-1'
    PaddingTopExtraLarge2 = 'pt-xl-2'
    PaddingTopExtraLarge3 = 'pt-xl-3'
    PaddingTopExtraLarge4 = 'pt-xl-4'
    PaddingTopExtraLarge5 = 'pt-xl-5'
    PaddingX0 = 'px-0'
    PaddingX1 = 'px-1'
    PaddingX2 = 'px-2'
    PaddingX3 = 'px-3'
    PaddingX4 = 'px-4'
    PaddingX5 = 'px-5'
    PaddingExtraLarge0 = 'p-xl-0'
    PaddingExtraLarge1 = 'p-xl-1'
    PaddingExtraLarge2 = 'p-xl-2'
    PaddingExtraLarge3 = 'p-xl-3'
    PaddingExtraLarge4 = 'p-xl-4'
    PaddingExtraLarge5 = 'p-xl-5'
    PaddingXLarge0 = 'px-lg-0'
    PaddingXLarge1 = 'px-lg-1'
    PaddingXLarge2 = 'px-lg-2'
    PaddingXLarge3 = 'px-lg-3'
    PaddingXLarge4 = 'px-lg-4'
    PaddingXLarge5 = 'px-lg-5'
    PaddingXMedium0 = 'px-md-0'
    PaddingXMedium1 = 'px-md-1'
    PaddingXMedium2 = 'px-md-2'
    PaddingXMedium3 = 'px-md-3'
    PaddingXMedium4 = 'px-md-4'
    PaddingXMedium5 = 'px-md-5'
    PaddingXSmall0 = 'px-sm-0'
    PaddingXSmall1 = 'px-sm-1'
    PaddingXSmall2 = 'px-sm-2'
    PaddingXSmall3 = 'px-sm-3'
    PaddingXSmall4 = 'px-sm-4'
    PaddingXSmall5 = 'px-sm-5'
    PaddingXExtraLarge0 = 'px-xl-0'
    PaddingXExtraLarge1 = 'px-xl-1'
    PaddingXExtraLarge2 = 'px-xl-2'
    PaddingXExtraLarge3 = 'px-xl-3'
    PaddingXExtraLarge4 = 'px-xl-4'
    PaddingXExtraLarge5 = 'px-xl-5'
    PaddingY0 = 'py-0'
    PaddingY1 = 'py-1'
    PaddingY2 = 'py-2'
    PaddingY3 = 'py-3'
    PaddingY4 = 'py-4'
    PaddingY5 = 'py-5'
    PaddingYLarge0 = 'py-lg-0'
    PaddingYLarge1 = 'py-lg-1'
    PaddingYLarge2 = 'py-lg-2'
    PaddingYLarge3 = 'py-lg-3'
    PaddingYLarge4 = 'py-lg-4'
    PaddingYLarge5 = 'py-lg-5'
    PaddingYMedium0 = 'py-md-0'
    PaddingYMedium1 = 'py-md-1'
    PaddingYMedium2 = 'py-md-2'
    PaddingYMedium3 = 'py-md-3'
    PaddingYMedium4 = 'py-md-4'
    PaddingYMedium5 = 'py-md-5'
    PaddingYSmall0 = 'py-sm-0'
    PaddingYSmall1 = 'py-sm-1'
    PaddingYSmall2 = 'py-sm-2'
    PaddingYSmall3 = 'py-sm-3'
    PaddingYSmall4 = 'py-sm-4'
    PaddingYSmall5 = 'py-sm-5'
    PaddingYExtraLarge0 = 'py-xl-0'
    PaddingYExtraLarge1 = 'py-xl-1'
    PaddingYExtraLarge2 = 'py-xl-2'
    PaddingYExtraLarge3 = 'py-xl-3'
    PaddingYExtraLarge4 = 'py-xl-4'
    PaddingYExtraLarge5 = 'py-xl-5'
    Rounded = 'rounded'
    Rounded0 = 'rounded-0'
    RoundedBottom = 'rounded-bottom'
    RoundedCircle = 'rounded-circle'
    RoundedLeft = 'rounded-left'
    RoundedLarge = 'rounded-lg'
    RoundedPill = 'rounded-pill'
    RoundedRight = 'rounded-right'
    RoundedSmall = 'rounded-sm'
    RoundedTop = 'rounded-top'
    Row = 'row'
    Shadow = 'shadow'
    ShadowLarge = 'shadow-lg'
    ShadowNone = 'shadow-none'
    ShadowSmall = 'shadow-sm'
    Show = 'show'
    Showing = 'showing'
    Small = 'small'
    SpinnerBorder = 'spinner-border'
    SpinnerBorderSmall = 'spinner-border-sm'
    SpinnerGrow = 'spinner-grow'
    SpinnerGrowSmall = 'spinner-grow-sm'
    SrOnly = 'sr-only'
    StickyTop = 'sticky-top'
    Table = 'table'
    TableActive = 'table-active'
    TableBordered = 'table-bordered'
    TableDanger = 'table-danger'
    TableDark = 'table-dark'
    TableInfo = 'table-info'
    TableLight = 'table-light'
    TablePrimary = 'table-primary'
    TableResponsive = 'table-responsive'
    TableResponsiveLarge = 'table-responsive-lg'
    TableResponsiveMedium = 'table-responsive-md'
    TableResponsiveSmall = 'table-responsive-sm'
    TableResponsiveExtraLarge = 'table-responsive-xl'
    TableSecondary = 'table-secondary'
    TableSuccess = 'table-success'
    TableWarning = 'table-warning'
    TabPane = 'tab-pane'
    TextBlack50 = 'text-black-50'
    TextBody = 'text-body'
    TextBreak = 'text-break'
    TextCapitalize = 'text-capitalize'
    TextCenter = 'text-center'
    TextDanger = 'text-danger'
    TextDark = 'text-dark'
    TextDecorationNone = 'text-decoration-none'
    TextHide = 'text-hide'
    TextInfo = 'text-info'
    TextJustify = 'text-justify'
    TextLeft = 'text-left'
    TextLargeCenter = 'text-lg-center'
    TextLargeLeft = 'text-lg-left'
    TextLargeRight = 'text-lg-right'
    TextLight = 'text-light'
    TextLowercase = 'text-lowercase'
    TextMediumCenter = 'text-md-center'
    TextMediumLeft = 'text-md-left'
    TextMediumRight = 'text-md-right'
    TextMonospace = 'text-monospace'
    TextMuted = 'text-muted'
    TextNowrap = 'text-nowrap'
    TextPrimary = 'text-primary'
    TextReset = 'text-reset'
    TextRight = 'text-right'
    TextSecondary = 'text-secondary'
    TextSmallCenter = 'text-sm-center'
    TextSmallLeft = 'text-sm-left'
    TextSmallRight = 'text-sm-right'
    TextSuccess = 'text-success'
    TextTruncate = 'text-truncate'
    TextUppercase = 'text-uppercase'
    TextWarning = 'text-warning'
    TextWhite = 'text-white'
    TextWhite50 = 'text-white-50'
    TextWrap = 'text-wrap'
    TextExtraLargeCenter = 'text-xl-center'
    TextExtraLargeLeft = 'text-xl-left'
    TextExtraLargeRight = 'text-xl-right'
    Toast = 'toast'
    ToastBody = 'toast-body'
    ToastHeader = 'toast-header'
    Tooltip = 'tooltip'
    TooltipInner = 'tooltip-inner'
    ValidFeedback = 'valid-feedback'
    ValidTooltip = 'valid-tooltip'
    FullHeight = 'vh-100'
    Visible = 'visible'
    FullWidth = 'vw-100'
    Width100 = 'w-100'
    Width25 = 'w-25'
    Width50 = 'w-50'
    Width75 = 'w-75'
    WidthAuto = 'w-auto'

    def __str__(self):
        return self.value

    def __call__(self, obj):
        return obj.add_class(self.value)


"""
Convenience classes for hooking into the matplotlib animation framework
"""
from .Graphics import GraphicsBase, Graphics

__all__ = [
    "EventHandler",
    "Animator"
    # "FrameAnimator"
]

class EventHandler:
    def __init__(self,
                 figure,
                 on_click = None,
                 on_release = None,
                 on_draw = None,
                 on_key_press = None,
                 on_key_release = None,
                 on_move = None,
                 on_select = None,
                 on_resize = None,
                 on_scroll = None,
                 on_figure_entered = None,
                 on_figure_left = None,
                 on_axes_entered = None,
                 on_axes_left = None
                 ):
        """Creates an EventHandler on a Figure that handles most interactivity stuff

        :param figure:
        :type figure: GraphicsBase
        :param on_click:
        :type on_click:
        :param on_release:
        :type on_release:
        :param on_draw:
        :type on_draw:
        :param on_key_press:
        :type on_key_press:
        :param on_key_release:
        :type on_key_release:
        :param on_move:
        :type on_move:
        :param on_select:
        :type on_select:
        :param on_resize:
        :type on_resize:
        :param on_scroll:
        :type on_scroll:
        :param on_figure_entered:
        :type on_figure_entered:
        :param on_figure_left:
        :type on_figure_left:
        :param on_axes_entered:
        :type on_axes_entered:
        :param on_axes_left:
        :type on_axes_left:
        """

        self.figure = figure

        if not isinstance(on_click, self.Event):
            on_click = self.ButtonPressedEvent(on_click)
        if not isinstance(on_release, self.Event):
            on_release = self.ButtonReleasedEvent(on_release)
        if not isinstance(on_draw, self.Event):
            on_draw = self.DrawEvent(on_draw)
        if not isinstance(on_key_press, self.Event):
            on_key_press = self.KeyPressedEvent(on_key_press)
        if not isinstance(on_key_release, self.Event):
            on_key_release = self.KeyReleasedEvent(on_key_release)
        if not isinstance(on_move, self.Event):
            on_move = self.MoveEvent(on_move)
        if not isinstance(on_select, self.Event):
            on_select = self.SelectEvent(on_select)
        if not isinstance(on_scroll, self.Event):
            on_scroll = self.ScrollEvent(on_scroll)
        if not isinstance(on_figure_entered, self.Event):
            on_figure_entered = self.FigureEnterEvent(on_figure_entered)
        if not isinstance(on_figure_left, self.Event):
            on_figure_left = self.FigureLeaveEvent(on_figure_left)
        if not isinstance(on_axes_entered, self.Event):
            on_axes_entered = self.AxesEnterEvent(on_axes_entered)
        if not isinstance(on_axes_left, self.Event):
            on_axes_left = self.AxesEnterEvent(on_axes_left)

        self._handles = dict(
            button_press_event = on_click,
            button_release_event =on_release,
            draw_event = on_draw,
            key_press_event = on_key_press,
            key_release_event = on_key_release,
            motion_notify_event = on_move,
            pick_event = on_select,
            resize_event = on_resize,
            scroll_event = on_scroll,
            figure_enter_event = on_figure_entered,
            figure_leave_event = on_figure_left,
            axes_enter_event = on_axes_entered,
            axes_leave_event = on_axes_left
        )

        self.bind(**self._handles)
    def bind(self, **handlers):
        self._handles.update(**handlers)
        for handle, handler in handlers.items():
            self.figure.canvas.mpl_connect(handle, handler)

    @property
    def handlers(self):
        proc_handle = lambda h: self._handles[h].data if isinstance(self._handles[h], self.Event) else None
        return {
            h: proc_handle(h) for h in self._handles
        }

    class Event:
        def __init__(self,
                     event_handler,
                     handler,
                     filter = None,
                     update = True,
                     name = None
                     ):
            self.filter = filter
            self.handler = handler
            self.parent = event_handler
            self.update = update,
            self.name = name

        @property
        def data(self):
            return (
                self.handler,
                {
                    'filter':self.filter,
                    'update':self.update,
                    'name':self.name
                    }
            )

        def handle_event(self, event):
            res = None
            if self.handler is not None:
                if self.filter is not None:
                    handle = self.filter(event)
                else:
                    handle = True
                if handle:
                    res = self.handler(event)
                    if (res is not False) and (res is not None or self.update):
                        self.parent.figure.canvas.draw()
            return res

        def __call__(self, *args, **kwargs):
            self.handle_event(*args, **kwargs)

    def ButtonPressedEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'ButtonPressed'}, **kw)
        return self.Event(self, handler, **kw)
    def ButtonReleasedEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'ButtonReleased'}, **kw)
        return self.Event(self, handler, **kw)
    def DrawEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'Draw'}, **kw)
        return self.Event(self, handler, **kw)
    def KeyPressedEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        if isinstance(handler, dict):
            filt = lambda e, h=handler: e.key in h
            handler = lambda e, h = handler: h[e.key](e)
        else:
            filt = None
        kw = dict({'name' : 'KeyPressed', 'filter':filt}, **kw)
        return self.Event(self, handler, **kw)
    def KeyReleasedEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        if isinstance(handler, dict):
            filt = lambda e, h=handler: e.key in h
            handler = lambda e, h = handler: h[e.key](e)
        else:
            filt = None
        kw = dict({'name' : 'KeyReleased', 'filter':filt}, **kw)
        return self.Event(self, handler, **kw)
    def MoveEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'Move'}, **kw)
        return self.Event(self, handler, **kw)
    def SelectEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'Select'}, **kw)
        return self.Event(self, handler, **kw)
    def ScrollEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'Scroll'}, **kw)
        return self.Event(self, handler, **kw)
    def FigureEnterEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'FigureEnter'}, **kw)
        return self.Event(self, handler, **kw)
    def FigureLeaveEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'FigureLeave'}, **kw)
        return self.Event(self, handler, **kw)
    def AxesEnterEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'AxesEnter'}, **kw)
        return self.Event(self, handler, **kw)
    def AxesLeaveEvent(self, handler, **kw):
        if isinstance(handler, tuple):
            handler, pars = kw
        else:
            pars = {}
        kw = dict(pars, **kw)
        kw = dict({'name' : 'AxesLeave'}, **kw)
        return self.Event(self, handler, **kw)

class Animator:
    def __init__(self, figure, data_generator,
                 plot_method = None,
                 events = True,
                 update = False,
                 **anim_ops
                 ):

        from matplotlib.animation import FuncAnimation
        from matplotlib.figure import Figure
        from functools import wraps

        self.figure = figure
        if isinstance(figure, Figure):
            figure = Graphics(figure=figure, axes=figure.axes)
        self.data_generator = data_generator
        self.plotter = self._get_plot_method(figure, plot_method)
        self.update = update
        self._prev_objects = None

        @wraps(self.plotter)
        def anim_func(frame, *args, self = self, **kwargs):
            try:
                if data_generator is not None:
                    data = self.data_generator(frame, *args)
                else:
                    data = args
            except (TypeError, ValueError):
                self.plotter = self.data_generator
                self.data_generator = None
                data = args

            if not self.update:
                if self._prev_objects is None:
                    self.figure.clear()
                else:
                    for o in self._prev_objects:
                        o.remove()

            if data is None:
                self.plotter = self.data_generator
                self.data_generator = None
            else:
                self._prev_objects = self.plotter(self.figure, *data, frame = frame, **kwargs)

        self._animation = FuncAnimation(figure.figure, anim_func, **anim_ops)
        self._active = True

        if events:
            self.figure.bind_events(
                on_click = lambda *e:self.toggle()
            )

    def _get_plot_method(self, figure, plot_method):
        if plot_method is None:
            plot_method = figure.plot
            plot_method = lambda figure, *data, m = plot_method, **kwargs: m(*data)

        return plot_method

    @property
    def active(self):
        return self._active
    @active.setter
    def active(self, val):
        if val:
            self.start()
        else:
            self.stop()
    def start(self):
        if not self.active:
            self._active = True
            self._animation.event_source.start()
    def stop(self):
        if self.active:
            self._active = False
            self._animation.event_source.stop()
    def toggle(self):
        if self.active:
            self.stop()
        else:
            self.start()
    def show(self):
        return self.figure.show()

    def to_jshtml(self):
        """
        Delegates to the underlying animation
        :return:
        :rtype:
        """
        return self._animation.to_jshtml()
    def to_html5_video(self):
        """
        Delegates to the underlying animation
        :return:
        :rtype:
        """
        return self._animation.to_html5_video()

    def as_jupyter_animation(self, mode='javascript'):
        """
        Chains some stuff to make Jupyter animations work
        :return:
        :rtype:
        """

        from IPython.display import HTML
        if mode == 'javascript':
            buffer = self.to_jshtml()
        else:
            buffer = self.to_html5_video()
        return HTML(buffer)

# class FrameAnimator:
#     def __init__(self, figure, *frame, **opts):
#         figure.animate_frames


from .AffineTransform import AffineTransform
import numpy as np

######################################################################################################
##
##                                   RotationTransform Class
##
######################################################################################################

__all__ = [
    "RotationTransform"
]

__reload_hook__ = [".AffineTransform"]

class RotationTransform(AffineTransform):
    """A simple AffineTransform implementation of the TransformationFunction abstract base class

    """

    def __init__(self, theta, axis="z", center = None):
        """

        :param theta: angle through which to rotate
        :type theta: float
        :param axis: axis about which to rotate
        :type axis: axis about which to rotate
        :param center: center point for the rotation
        :type center: None or np.array
        """

        from ...Numputils import rotation_matrix, translation_matrix, merge_transformation_matrix

        rot_mat = rotation_matrix(axis, theta)
        if center is not None:
            translate = translation_matrix(-center)
            neg_translate = translation_matrix(center)
            rot_mat = merge_transformation_matrix(merge_transformation_matrix(translate, rot_mat), neg_translate)

        super().__init__(rot_mat, shift=None)

    def reverse(self):
        base_mat = self.transform
        shift = self.shift
        inverse = base_mat.transpose()
        new_shift = -np.dot(inverse, shift)
        type(self)(inverse, shift=new_shift)




__all__ = []
from .Variables import *; from .Variables import __all__ as exposed
__all__ += exposed
from .Interfaces import *; from .Interfaces import __all__ as exposed
__all__ += exposed
from .Controls import *; from .Controls import __all__ as exposed
__all__ += exposed
from .Apps import *; from .Apps import __all__ as exposed
__all__ += exposed


import numpy as np
from .TransformationFunction import TransformationFunction
from ...Numputils import affine_matrix, merge_transformation_matrix, one_pad_vecs

######################################################################################################
##
##                                   AffineTranform Class
##
######################################################################################################

__all__ = [
    "AffineTransform"
]

__reload_hook__ = ['.TransformationFunction', "...Numputils"]

class AffineTransform(TransformationFunction):
    """A simple AffineTranform implementation of the TransformationFunction abstract base class

    """

    def __init__(self, tmat, shift=None):
        """tmat must be a transformation matrix to work properly

        :param shift: the shift for the transformation
        :type shift: np.ndarray | None
        :param tmat: the matrix for the linear transformation
        :type tmat: np.ndarray
        """

        self.transf = affine_matrix(tmat, shift)
        super().__init__()

    @property
    def transform(self):
        return self.transf[:3, :3]

    @property
    def inverse(self):
        """
        Returns the inverse of the transformation
        :return:
        :rtype:
        """
        return self.reverse()

    @property
    def shift(self):
        transf = self.transf
        transf_shape = transf.shape
        if transf_shape[-1] == 4:
            vec = transf[:-1, 3] # hopefully this copies rather than just being a view...
        else:
            vec = None
        return vec

    def merge(self, other):
        """

        :param other:
        :type other: np.ndarray or AffineTransform
        """

        if isinstance(other, AffineTransform):
            other = other.transf
        transf = self.transf

        # I wanted to use type(self) but then I realized that'll fuck me over
        # if I want to merge like a ScalingTransform and a RotationTransform
        return AffineTransform(merge_transformation_matrix(transf, other))


    def reverse(self):
        """Inverts the matrix

        :return:
        :rtype:
        """
        
        inverse = np.linalg.inv(self.transf)
        return AffineTransform(inverse)


    def operate(self, coords, shift=True):
        """

        :param coords: the array of coordinates passed in
        :type coords: np.ndarry
        """
        translate = shift

        # Assumes that we're getting 3D cartesian coordinates...might not be a valid assumption
        coords = np.asarray(coords)
        coord_shape = coords.shape
        if len(coord_shape) == 1:
            adj_coord = coords.reshape((1, coord_shape[0]))
        elif len(coord_shape) > 2:
            nels = np.prod(coord_shape[:-1])
            adj_coord = coords.reshape((nels, 3))
        else:
            adj_coord = coords

        tmat = self.transf
        if tmat.shape[-1] == 4:
            if not translate:
                tmat = tmat[:3, :3]
            else:
                adj_coord = one_pad_vecs(adj_coord)
            adj_coord = np.tensordot(adj_coord, tmat, axes=[1, 1])
            if translate:
                adj_coord = adj_coord[..., :3]
            adj_coord = adj_coord.reshape(coord_shape)
        else:
            adj_coord = np.tensordot(adj_coord, tmat, axes=[1, 1])
            adj_coord = adj_coord.reshape(coord_shape)

        return adj_coord

    def __repr__(self):
        ## we'll basically just leverage the ndarray repr:
        return "{}(transformation={}, shift={})".format(type(self).__name__, str(self.transform), str(self.shift))


from .CoordinateSystemConverter import CoordinateSystemConverter
from .CommonCoordinateSystems import CartesianCoordinates3D, ZMatrixCoordinates
from ...Numputils import vec_norms, vec_angles, vec_dihedrals, pts_dihedrals, dist_deriv, angle_deriv, dihed_deriv, find
from ...Scaffolding import MaxSizeCache
import numpy as np
# this import gets bound at load time, so unfortunately PyCharm can't know just yet
# what properties its class will have and will try to claim that the files don't exist

class CartesianToZMatrixConverter(CoordinateSystemConverter):
    """
    A converter class for going from Cartesian coordinates to ZMatrix coordinates
    """

    @property
    def types(self):
        return (CartesianCoordinates3D, ZMatrixCoordinates)

    @staticmethod
    def get_dists(points, centers, return_diffs=False):
        diffs = centers-points
        ret = vec_norms(diffs)
        if return_diffs:
            ret = (ret, diffs)
        return ret
    @staticmethod
    def get_angles(lefts, centers, rights, norms=None):
        # need to look up again what the convention is for which atom is the central one...
        if rights is None:
            v1s = lefts
            v2s = centers
        else:
            v1s = centers-lefts
            v2s = centers-rights
        return vec_angles(v1s, v2s, norms=norms, return_crosses=True, return_norms=False)
    @staticmethod
    def get_diheds(points, centers, seconds, thirds, crosses=None, norms=None):
        if thirds is None:
            return vec_dihedrals(points, centers, seconds,
                                 crosses=crosses,
                                 norms=norms
                                 )
        else:
            return pts_dihedrals(points, centers, seconds, thirds, crosses=crosses, norms=norms)

    def convert_many(self, coords, ordering=None, use_rad=True, return_derivs=False, **kw):
        """
        We'll implement this by having the ordering arg wrap around in coords?
        """
        if ordering is None:
            ordering = range(len(coords[0]))
        base_shape = coords.shape
        new_coords = np.reshape(coords, (np.prod(base_shape[:-1]),) + base_shape[-1:])
        new_coords, ops = self.convert(new_coords, ordering=ordering, use_rad=use_rad, return_derivs=return_derivs)
        single_coord_shape = (base_shape[-2]-1, new_coords.shape[-1])
        new_shape = base_shape[:-2] + single_coord_shape
        new_coords = np.reshape(new_coords, new_shape)
        if return_derivs:
            ders = ops['derivs']
            # we assume we get a list of derivatives?
            reshaped_ders = [None]*len(ders)
            for i, v in enumerate(ders):
                single_base_shape = (base_shape[-2], new_coords.shape[-1])
                ders_shape = coords.shape + single_base_shape*i + single_coord_shape
                v = v.reshape(ders_shape)
                reshaped_ders[i] = v
                ops['derivs'] = reshaped_ders
        return new_coords, ops

    @staticmethod
    def _get_conversion_info(ncoords, ordering, canonicalize):
        if canonicalize:
            orig_ol = ZMatrixCoordinates.canonicalize_order_list(ncoords, ordering)
        else:
            orig_ol = ordering
        ol = orig_ol
        nol = len(ol)
        ncol = len(ol[0])

        return ol, nol, ncol

    @staticmethod
    def _build_ordering_list(multiconfig, ol, ncoords):
        if multiconfig:
            ol = ZMatrixCoordinates.tile_order_list(ol, ncoords)
            mc_ol = ol.copy()
        else:
            mc_ol = None

        # we define an order map that we'll index into to get the new indices for a
        # given coordinate
        om = 1+np.argsort(ol[:, 0])

        return ol, om, mc_ol

    @staticmethod
    def _get_dist_inds(mask, ol, nol):
        subs = ol.reshape(-1, nol, 4)[:, 1:].reshape(-1, 4)
        # subs = ol[mask]
        ix = subs[:, 0]
        jx = subs[:, 1]
        return ix, jx
    def _get_cached_dist_inds(self, ol_key, ncoords, ol, nol):
        if (ol_key, ncoords) in self._index_cache:
            idx_list = self._index_cache[(ol_key, ncoords)]
        else:
            # mask = np.arange(ncoords).reshape(-1, nol)[:, 1:].flatten()
            # mask = np.setdiff1d(np.arange(ncoords), np.arange(0, ncoords, nol))
            # raise Exception((mask_1 == mask).all())
            ix, jx = self._get_dist_inds(None, ol, nol)
            idx_list = [[ix, jx]]
            self._index_cache[(ol_key, ncoords)] = idx_list
        return idx_list

    def _get_cached_norms(self, distance_matrix, coords, ix, jx):
        target_norms_1 = distance_matrix[ix, jx]
        bad_norms_1 = np.where(target_norms_1 == 0)
        if len(bad_norms_1) > 0 and len(bad_norms_1[0]) > 0:
            bad_ix = ix[bad_norms_1]
            bad_jx = jx[bad_norms_1]
            bad_dists = self.get_dists(coords[bad_ix], coords[bad_jx])
            distance_matrix[bad_ix, bad_jx] = bad_dists
            distance_matrix[bad_jx, bad_ix] = bad_dists
            target_norms_1[bad_norms_1] = distance_matrix[bad_ix]
        return target_norms_1

    @staticmethod
    def _get_angle_inds(mask, ol, nol):
        subs = ol.reshape(-1, nol, 4)[:, 2:].reshape(-1, 4)
        # subs = ol[mask]
        ix = subs[:, 0]
        jx = subs[:, 1]
        kx = subs[:, 2]

        return ix, jx, kx
    def _get_cached_angle_inds(self, idx_list, mask, ncoords, ol, nol):
        if len(idx_list) < 2:
            # mask = mask.reshape(-1, nol-1)[:, 1:].flatten()
            # mask = np.delete(mask, np.arange(0, len(mask), nol-1))
            # mask = np.setdiff1d(mask, np.arange(1, ncoords, nol))
            # raise Exception(np.all(mask_1 == mask))
            ix, jx, kx = self._get_angle_inds(None, ol, nol)
            idx_list.append([ix, jx, kx])
        return idx_list

    def _get_angles_cached(self, distance_matrix, cross_mat, cross_norms, coords, ix, jx, kx):
        norms_1 = self._get_cached_norms(distance_matrix, coords, ix, jx)
        norms_2 = self._get_cached_norms(distance_matrix, coords, kx, jx)
        angles, crosses = self.get_angles(
            coords[ix], coords[jx], coords[kx],
            norms=[norms_1, norms_2]
        )
        cross_mat[ix, jx, kx] = crosses
        cross_mat[kx, jx, ix] = -crosses
        cnorms = vec_norms(crosses)
        cross_norms[ix, jx, kx] = cnorms
        cross_norms[kx, ix, jx] = cnorms
        return angles
    def _get_cached_crosses(self, distance_matrix, cross_mat, cross_norms, coords, ix, jx, kx):
        target_norms = cross_norms[ix, jx, kx]
        bad_norms = np.where(target_norms == 0)
        if len(bad_norms) > 0 and len(bad_norms[0]) > 0:
            bad_ix = ix[bad_norms]
            bad_jx = jx[bad_norms]
            bad_kx = kx[bad_norms]
            self._get_angles_cached(distance_matrix, cross_mat, cross_norms, coords, bad_ix, bad_jx, bad_kx)
        return cross_mat[ix, jx, kx], cross_norms[ix, jx, kx] # recalculated in the angles_cached call

    @staticmethod
    def _get_dihed_inds(mask, ol, nol):
        subs = ol.reshape(-1, nol, 4)[:, 3:].reshape(-1, 4)
        ix = subs[:, 0]
        jx = subs[:, 1]
        kx = subs[:, 2]
        lx = subs[:, 3]

        return ix, jx, kx, lx
    def _get_cached_dihed_inds(self, idx_list, mask, ncoords, ol, nol):
        if len(idx_list) < 3:
            # mask = np.setdiff1d(mask, np.arange(2, ncoords, nol))
            # mask = np.delete(mask, np.arange(0, len(mask), nol-2))
            # mask = mask.reshape(-1, nol-2)[:, 1:].flatten()
            ix, jx, kx, lx = self._get_dihed_inds(None, ol, nol)
            idx_list.append([ix, jx, kx, lx])
        return idx_list

    @staticmethod
    def _restructure_angles(angles, ncoords, steps, nol, use_rad):
        angles = np.append(angles, np.zeros(steps))
        insert_pos = np.arange(0, ncoords - 1 * steps - 1, nol - 2)
        angles = np.insert(angles, insert_pos, 0)
        angles = angles[:ncoords - steps]
        if not use_rad:
            angles = np.rad2deg(angles)
        return angles

    @staticmethod
    def _restructure_diheds(diheds, ncoords, steps, nol, use_rad):
        # pad diheds to be the size of ncoords
        diheds = np.append(diheds, np.zeros(2 * steps))

        # insert zeros where undefined
        diheds = np.insert(diheds, np.repeat(np.arange(0, ncoords - 2 * steps - 1, nol - 3), 2), 0)
        # take only as many as actually used
        diheds = diheds[:ncoords - steps]
        if not use_rad:
            diheds = np.rad2deg(diheds)
        return diheds

    @staticmethod
    def _build_post_ordering_lists(ol, om, ncoords, nol, ncol, steps):
        mask = np.full(ncoords, True)
        mask[np.arange(0, ncoords, nol)] = False
        ol = np.reshape(ol[mask], (steps, nol - 1, ncol)) - np.reshape(np.arange(steps), (steps, 1, 1))
        ol = np.reshape(ol, (ncoords - steps, ncol))
        om = np.reshape(om[mask], (steps, nol - 1)) - nol * np.reshape(np.arange(steps), (steps, 1)) - 1
        om = np.reshape(om, (ncoords - steps,))
        return ol, om

    @staticmethod
    def _build_coord_triples(dists, angles, diheds):
        return np.concatenate(
            [
                dists.reshape(-1, 1),
                angles.reshape(-1, 1),
                diheds.reshape(-1, 1)
            ],
            axis=-1
        )

    @staticmethod
    def _rebuild_ordering(ncol, ol, om):
        om = om - 1
        if ncol == 5:
            ordering = np.array([
                        np.argsort(ol[:, 0]), om[ol[:, 1]], om[ol[:, 2]], om[ol[:, 3]], ol[:, 4]
                    ]).T
        else:
            ordering = np.array([
                    np.argsort(ol[:, 0]), om[ol[:, 1]], om[ol[:, 2]], om[ol[:, 3]]
                ]).T
        return ordering

    @staticmethod
    def _compute_final_frame(multiconfig, coords, mc_ol, nol, ol):
        if multiconfig:
            # figure out what to use for the axes
            origins = coords[mc_ol[1::nol,  1]]
            x_axes  = coords[mc_ol[1::nol,  0]] - origins # the first displacement vector
            y_axes  = coords[mc_ol[2::nol,  0]] - origins # the second displacement vector (just defines the x-y plane, not the real y-axis)
            axes = np.array([x_axes, y_axes]).transpose((1, 0, 2))
        else:
            origins = coords[ol[0, 1]]
            axes = np.array([coords[ol[0, 0]] - origins, coords[ol[1, 0]] - origins])
        return origins, axes

    def _convert_multiconfig(self, coords, ol, om, orig_ol, ncoords, nol, ncol, steps, use_rad, return_derivs, derivs):

        # we do all of this stuff with masking operations in the multiconfiguration cases
        # ...except the mask was too slow so now we don't...
        if return_derivs:
            raise NotImplementedError(...)

        ol_key = tuple(tuple(o) for o in orig_ol)
        idx_list = self._get_cached_dist_inds(ol_key, ncoords, ol, nol)
        ix, jx = idx_list[0]
        # dupe over number of structures
        dist_pos = [[ix, jx]]
        num_dists = len(ix)
        if nol > 2:
            idx_list = self._get_cached_angle_inds(idx_list, None, ncoords, ol, nol)
            ix, jx, kx = idx_list[1] # mask
            num_angles = len(ix)

            dist_pos = dist_pos + [[jx, kx]]
            cross_pos = [[ix, jx, kx]]
            if nol > 3:
                idx_list = self._get_cached_dihed_inds(idx_list, None, ncoords, ol, nol)
                # mask[np.arange(2, ncoords, nol)] = 0
                ix, jx, kx, lx = idx_list[2]
                # raise Exception(
                #     [ix, dist_pos[0][0]],
                #     [jx, dist_pos[0][1]]
                # )
                if ol.shape[1] == 5:
                    raise ValueError("Unclear if there is a difference between tau and psi")
                    ix = ix.copy()
                    jx = jx.copy()
                    kx = kx.copy()
                    lx = lx.copy()
                    fx = ol[mask, 4]
                    swap_pos = np.where(fx == 1)
                    swap_i = ix[swap_pos]
                    swap_j = jx[swap_pos]
                    swap_k = kx[swap_pos]
                    swap_l = lx[swap_pos]
                    ix[swap_pos] = swap_l
                    jx[swap_pos] = swap_i
                    kx[swap_pos] = swap_j
                    lx[swap_pos] = swap_k
                dist_pos = dist_pos + [[kx, lx]]
                if len(idx_list) < 4:
                    all_dist_inds = np.concatenate(
                            [
                                np.concatenate([x[:, np.newaxis] for x in dlist], axis=-1)
                                for dlist in dist_pos
                            ],
                            axis=0
                        )
                    ud_inds, ud_inv = np.unique(
                        all_dist_inds,
                        axis=0,
                        # return_index=True,
                        return_inverse=True
                    )
                    # ud_sorting = np.argsort(ud_pos)
                    # ud_inv_sort = np.argsort(ud_sorting)
                    # ud_pos = ud_pos[ud_sorting]
                    # ud_inv = ud_inv_sort[ud_inv]
                    idx_list.append([ud_inds, ud_inv])

                cross_pos = cross_pos + [[jx, kx, lx]]
                if len(idx_list) < 5:
                    all_cross_pos = np.concatenate(
                        [
                            np.concatenate([x[:, np.newaxis] for x in clist], axis=-1)
                            for clist in cross_pos
                        ],
                        axis=0
                    )

                    # determine which unique cross products we need to calculate
                    # and make sure they remain sorted so we can just take the first chunks
                    uc_inds, uc_inv = np.unique(
                        all_cross_pos,
                        axis=0,
                        # return_index=True,
                        return_inverse=True
                    )

                    # uc_sorting = np.argsort(uc_pos)
                    # uc_inv_sort = np.argsort(uc_sorting)
                    # uc_pos = uc_pos[uc_sorting]
                    # uc_inv = uc_inv_sort[uc_inv]

                    # now we map these onto the corresponding distance positions

                    # raise Exception(
                    #     all_dist_inds[:5],
                    #     ud_inds_sorted[:5],
                    #     all_cross_pos[:5, (0, 1)]
                    # )
                    v1_map, _ = find(
                        ud_inds,
                        uc_inds[:, (0, 1)],
                        sorting='sorted',
                        search_space_sorting=[
                            np.arange(len(uc_inds)),
                            np.arange(len(uc_inds))
                        ]
                    )
                    v2_map, _ = find(
                        ud_inds,
                        uc_inds[:, (1, 2)],
                        sorting='sorted',
                        search_space_sorting=None
                    )

                    idx_list.append([uc_inds, uc_inv, v1_map, v2_map])
                # dihed_pos = idx_list[2]
            else:
                raise NotImplementedError("need to manage sorting")
                if len(idx_list) < 3:
                    ud_inds, ud_inv = np.unique(
                        np.concatenate(
                            [
                                np.concatenate([x[:, np.newaxis] for x in dlist], axis=-1)
                                for dlist in dist_pos
                            ],
                            axis=0
                        ),
                        axis=0,
                        return_inverse=True
                    )
                    idx_list.append([ud_inds, ud_inv])
                if len(idx_list) < 4:
                    uc_inds, uc_pos, uc_inv = np.unique(
                        np.concatenate(
                            [
                                np.concatenate([x[:, np.newaxis] for x in clist], axis=-1)
                                for clist in cross_pos
                            ],
                            axis=0
                        ),
                        axis=0,
                        return_index=True,
                        return_inverse=True
                    )
                    idx_list.append([uc_inds, uc_pos, uc_inv])

        else:
            raise NotImplementedError("need to manage sorting")
            if len(idx_list) < 2:
                ud_inds, ud_inv = np.unique(
                    np.concatenate(
                        [
                            np.concatenate([x[:, np.newaxis] for x in dlist], axis=-1)
                            for dlist in dist_pos
                        ],
                        axis=0
                    ),
                    axis=0,
                    return_inverse=True
                )
                idx_list.append([ud_inds, ud_inv])
            if len(idx_list) < 3:
                idx_list.append(None)

        # distance_matrix = np.zeros((ncoords, ncoords)) # we track this for reuse throughout remaining routines
        ud_inds, ud_inv = idx_list[-2]
        # ud_inds = all_dist_inds[ud_pos]
        #np.unique(dist_pos, axis=0, return_inverse=True)
        ix, jx = ud_inds[:, 0], ud_inds[:, 1]
        unique_dists, unique_diffs = self.get_dists(coords[ix], coords[jx], return_diffs=True)
        dists = unique_dists[ud_inv[:num_dists]]
        # all_dists, all_diffs = self.get_dists(coords[ix], coords[jx], return_diffs=True)
        # dists = all_dists[:num_dists]
        if return_derivs:
            raise NotImplementedError("needs to be updated to account for uniqueness")
            _, dist_derivs, dist_derivs_2 = dist_deriv(coords, ix, jx, order=2)
            drang = np.arange(nol - 1)
            nreps = int(len(ix) / (nol - 1))
            drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
            derivs[0][ix, :, drang, 0] = dist_derivs[0]
            derivs[0][jx, :, drang, 0] = dist_derivs[1]

            for i, x1 in enumerate([ix, jx]):
                for j, x2 in enumerate([ix, jx]):
                    derivs[1][x1, :, x2 % nol, :, drang, 0] = dist_derivs_2[i, j]

        if nol > 2:
            # uc_inds, uc_pos, uc_inv = np.unique(cross_pos, axis=0, return_index=True, return_inverse=True)
            uc_inds, uc_inv, uc_norm_inds1, uc_norm_inds2 = idx_list[-1]

            # we now need to map the unique cross product indices to the corresponding
            # distance indices...
            # we do this by noting that if uc_pos < num_angles, there's a corresponding distance
            # index

            # indices used in computing the distances that are also used when computing the
            # angles and their corresponding norms
            # uc_norm_inds1 = angle_dist_inds[uc_pos]
            # uc_norm_inds2 = ud_inv[num_dists:num_dists+num_angles][uc_pos]

            # raise Exception(len(uc_norm_inds1), len(uc_norm_inds2))
            norms_1 = unique_dists[uc_norm_inds1]#.reshape(-1, nol-1)[:, 1:].flatten()
            diffs_1 = unique_diffs[uc_norm_inds1]#.reshape(-1, nol-1, 3)[:, 1:].reshape(-1, 3)
            norms_2 = unique_dists[uc_norm_inds2]
            diffs_2 = unique_diffs[uc_norm_inds2]

            # ix, jx, kx = uc_inds[:, 0], uc_inds[:, 1], uc_inds[:, 2]
            unique_angles, unique_crosses = self.get_angles(
                diffs_2,
                diffs_1, # flipping the order is the same as negating diffs_1 (which is otherwise necessary)
                None,
                norms=[norms_2, norms_1]
            )
            unique_cross_norms = vec_norms(unique_crosses)

            # idx_list, mask = self._get_cached_angle_inds(idx_list, mask, ncoords, ol, nol)
            # ix, jx, kx = idx_list[1]
            #
            #
            # cross_norms = np.zeros((ncoords, ncoords, ncoords))
            # cross_mat = np.zeros((ncoords, ncoords, ncoords, 3)) # we track this for reuse throughout remaining routines
            # angles = self._get_angles_cached(distance_matrix, cross_mat, cross_norms, coords, ix, jx, kx)
            angles = unique_angles[uc_inv[:num_angles]]
            # angles = all_angles[:num_angles]
            angles = self._restructure_angles(angles, ncoords, steps, nol, use_rad)
            if return_derivs:
                raise NotImplementedError("needs to be updated to account for uniqueness")
                # we might need to mess with the masks akin to the insert call...
                _, angle_derivs, angle_derivs_2 = angle_deriv(coords, jx, ix, kx, order=2)
                drang = 1 + np.arange(nol - 2)
                nreps = int(len(ix) / (nol - 2))
                drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                derivs[0][jx, :, drang, 1] = angle_derivs[0]
                derivs[0][ix, :, drang, 1] = angle_derivs[1]
                derivs[0][kx, :, drang, 1] = angle_derivs[2]

                for i, x1 in enumerate([ix, jx, kx]):
                    for j, x2 in enumerate([ix, jx, kx]):
                        derivs[1][x1, :, x2 % nol, :, drang, 0] = angle_derivs_2[i, j]
        else:
            angles = np.zeros(ncoords - steps)

        if nol > 3:
            norm_inds1 = ud_inv[:num_dists].reshape(-1, nol - 1)[:, 2:].flatten()
            norm_inds2 = ud_inv[num_dists:num_dists+num_angles].reshape(-1, nol - 2)[:, 1:].flatten()
            norm_inds3 = ud_inv[num_dists+num_angles:]

            norms_1 = unique_dists[norm_inds1]
            diffs_1 = unique_diffs[norm_inds1]
            norms_2 = unique_dists[norm_inds2]
            diffs_2 = unique_diffs[norm_inds2] # 3 - 2
            norms_3 = unique_dists[norm_inds3]
            diffs_3 = unique_diffs[norm_inds3] # 4 - 3

            cross_inds1 = uc_inv[:num_angles].reshape(-1, nol - 2)[:, 1:].flatten() # 3 - 2 x 2 - 1
            cross_inds2 = uc_inv[num_angles:] # 4 - 3 x 3 - 2
            crosses_1 = unique_crosses[cross_inds1]
            crosses_2 = unique_crosses[cross_inds2]
            cross_norms1 = unique_cross_norms[cross_inds1]
            cross_norms2 = unique_cross_norms[cross_inds2]

            diheds = self.get_diheds(
                diffs_1,
                diffs_2,
                diffs_3,
                None,
                norms=[norms_1, norms_2, norms_3],
                crosses=[[crosses_1, crosses_2], [cross_norms1, cross_norms2]]
            )
            diheds = self._restructure_diheds(diheds, ncoords, steps, nol, use_rad)

            # # set up mask to drop all of the second atom bits (wtf it means 'second')
            # idx_list, mask = self._get_cached_dihed_inds(idx_list, mask, ncoords, ol, nol)
            # # mask[np.arange(2, ncoords, nol)] = 0
            # ix, jx, kx, lx = idx_list[2]
            #
            # # norms_1 = self._get_cached_norms(distance_matrix, coords, ix, jx)
            # norms_2 = self._get_cached_norms(distance_matrix, coords, jx, kx)
            # # norms_3 = self._get_cached_norms(distance_matrix, coords, kx, lx)
            #
            # crosses_1, cross_norms_1 = self._get_cached_crosses(distance_matrix, cross_mat, cross_norms, coords, ix, jx, kx)
            # crosses_2, cross_norms_2 = self._get_cached_crosses(distance_matrix, cross_mat, cross_norms, coords, ix, jx, kx)
            #
            # diheds = self.get_diheds(coords[ix], coords[jx], coords[kx], coords[lx],
            #                          crosses=[(crosses_1, crosses_2), (cross_norms_1, cross_norms_2)],
            #                          norms=[None, norms_2, None] # all we really need to reuse...
            #                          )
            # diheds = self._restructure_diheds(diheds, ncoords, steps, nol, use_rad)

            # diheds = self.get_diheds(coords[ix], coords[jx], coords[kx], coords[lx],
            #                          crosses=[(crosses_1, crosses_2), (cross_norms_1, cross_norms_2)],
            #                          norms=[None, norms_2, None] # all we really need to reuse...
            #                          )

            if return_derivs:
                # Negative sign because my dihed_deriv code is for slightly different
                # ordering than expected
                _, dihed_derivs, dihed_derivs_2 = dihed_deriv(coords, ix, jx, kx, lx, order=2)
                drang = 2 + np.arange(nol - 3)
                nreps = int(len(ix) / (nol - 3))
                drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                derivs[0][ix, :, drang, 2] = dihed_derivs[0]
                derivs[0][jx, :, drang, 2] = dihed_derivs[1]
                derivs[0][kx, :, drang, 2] = dihed_derivs[2]
                derivs[0][lx, :, drang, 2] = dihed_derivs[3]

                for i, x1 in enumerate([ix, jx, kx, lx]):
                    for j, x2 in enumerate([ix, jx, kx, lx]):
                        derivs[1][x1, :, x2 % nol, :, drang, 0] = dihed_derivs_2[i, j]

        else:
            diheds = np.zeros(ncoords - steps)

        # after the np.insert calls we have the right number of final elements, but too many
        # ol and om elements and they're generally too large
        # so we need to shift them down and mask out the elements we don't want
        ol, om = self._build_post_ordering_lists(ol, om, ncoords, nol, ncol, steps)

        return dists, angles, diheds, ol, om


    _index_cache = MaxSizeCache()
    def convert(self, coords, ordering=None, use_rad=True, return_derivs=False, canonicalize=True, **kw):
        """The ordering should be specified like:

        [
            [n1],
            [n2, n1]
            [n3, n1/n2, n1/n2]
            [n4, n1/n2/n3, n1/n2/n3, n1/n2/n3]
            [n5, ...]
            ...
        ]

        :param coords:    array of cartesian coordinates
        :type coords:     np.ndarray
        :param use_rad:   whether to user radians or not
        :type use_rad:    bool
        :param ordering:  optional ordering parameter for the z-matrix
        :type ordering:   None or tuple of ints or tuple of tuple of ints
        :param kw:        ignored key-word arguments
        :type kw:
        :return: z-matrix coords
        :rtype: np.ndarray
        """
        ncoords = len(coords)
        ol, nol, ncol = self._get_conversion_info(ncoords, ordering, canonicalize)
        orig_ol = ol
        steps = ncoords // nol
        # steps = int(fsteps)

        # print(">> c2z >> ordering:", ol)

        multiconfig = nol < ncoords
        ol, om, mc_ol = self._build_ordering_list(multiconfig, ol, ncoords)

        # need to check against the cases of like 1, 2, 3 atom molecules
        # annoying but not hard
        if return_derivs:
            derivs = [
                np.zeros(coords.shape + (nol-1, 3)),
                np.zeros(coords.shape + (nol, 3) + (nol - 1, 3))
            ]
        else:
            derivs = None

        if not multiconfig:
            ix = ol[1:, 0]
            jx = ol[1:, 1]
            dists = self.get_dists(coords[ix], coords[jx])
            if return_derivs:
                _dists, dist_derivs, dist_derivs_2 = dist_deriv(coords, ix, jx, order=2)
                drang = np.arange(len(ix))
                derivs[0][ix, :, drang, 0] = dist_derivs[0]
                derivs[0][jx, :, drang, 0] = dist_derivs[1]

                for i, x1 in enumerate([ix, jx]):
                    for j, x2 in enumerate([ix, jx]):
                        # print(i, j, x1, x2,
                        #       # dist_derivs_2[i, j][0, 0],
                        #       drang
                        #       )
                        derivs[1][x1, :, x2, :, drang, 0] = dist_derivs_2[i, j]

            if len(ol) > 2:
                ix = ol[2:, 0]
                jx = ol[2:, 1]
                kx = ol[2:, 2]
                angles = np.concatenate( (
                    [0], self.get_angles(coords[ix], coords[jx], coords[kx])[0]
                ) )
                if not use_rad:
                    angles = np.rad2deg(angles)
                if return_derivs:
                    _angles, angle_derivs, angle_derivs_2 = angle_deriv(coords, jx, ix, kx, order=2)
                    drang = 1+np.arange(len(ix))
                    # print(">>>>", np.max(np.abs(angle_derivs)))
                    derivs[0][jx, :, drang, 1] = angle_derivs[0]
                    derivs[0][ix, :, drang, 1] = angle_derivs[1]
                    derivs[0][kx, :, drang, 1] = angle_derivs[2]

                    for i, x1 in enumerate([jx, ix, kx]):
                        for j, x2 in enumerate([jx, ix, kx]):
                            derivs[1][x1, :, x2, :, drang, 1] = angle_derivs_2[i, j]
            else:
                angles = np.array([0.])
            if len(ol) > 3:
                ix = ol[3:, 0]
                jx = ol[3:, 1]
                kx = ol[3:, 2]
                lx = ol[3:, 3]
                if ol.shape[1] == 5:
                    raise NotImplementedError("psi angles might be unnecessary")
                    ix = ix.copy()
                    jx = jx.copy()
                    kx = kx.copy()
                    lx = lx.copy()
                    fx = ol[3:, 4]
                    swap_pos = np.where(fx == 1)
                    swap_i = ix[swap_pos]
                    swap_j = jx[swap_pos]
                    swap_k = kx[swap_pos]
                    swap_l = lx[swap_pos]
                    ix[swap_pos] = swap_l
                    jx[swap_pos] = swap_i
                    kx[swap_pos] = swap_j
                    lx[swap_pos] = swap_k

                diheds = np.concatenate( (
                    [0, 0],
                    self.get_diheds(coords[ix], coords[jx], coords[kx], coords[lx])
                ) )
                if not use_rad:
                    diheds = np.rad2deg(diheds)
                if return_derivs:
                    _diheds, dihed_derivs, dihed_derivs_2 = dihed_deriv(coords, ix, jx, kx, lx, order=2)
                    drang = 2+np.arange(len(ix))
                    derivs[0][ix, :, drang, 2] = dihed_derivs[0]
                    derivs[0][jx, :, drang, 2] = dihed_derivs[1]
                    derivs[0][kx, :, drang, 2] = dihed_derivs[2]
                    derivs[0][lx, :, drang, 2] = dihed_derivs[3]

                    for i, x1 in enumerate([ix, jx, kx, lx]):
                        for j, x2 in enumerate([ix, jx, kx, lx]):
                            derivs[1][x1, :, x2, :, drang, 2] = dihed_derivs_2[i, j]
            else:
                diheds = np.array([0, 0])
            ol = ol[1:]

        else: # multiconfig
            dists, angles, diheds, ol, om = self._convert_multiconfig(
                coords,
                ol, om, orig_ol,
                ncoords, nol, ncol, steps,
                use_rad,
                return_derivs, derivs
            )

        final_coords = self._build_coord_triples(dists, angles, diheds)

        origins, axes = self._compute_final_frame(multiconfig, coords, mc_ol, nol, ol)

        ordering = self._rebuild_ordering(ncol, orig_ol, om)
        opts = dict(use_rad=use_rad, ordering=ordering, origins=origins, axes=axes)

        # if we're returning derivs, we also need to make sure that they're ordered the same way the other data is...
        if return_derivs:
            opts['derivs'] = derivs#[:1]

        return final_coords, opts

__converters__ = [ CartesianToZMatrixConverter() ]


"""
Handles much of the "numerical math" stuff inside Mcutils which has made it balloon a little bit
Deals with anything tensor, Taylor expansion, or interpolation related
"""

__all__ = []
from .Taylor import *; from .Taylor import __all__ as exposed
__all__ += exposed
from .Mesh import *; from .Mesh import __all__ as exposed
__all__ += exposed
from .Surfaces import *; from .Surfaces import __all__ as exposed
__all__ += exposed
from .FittableModels import *; from .FittableModels import __all__ as exposed
__all__ += exposed
from .Interpolator import *; from .Interpolator import __all__ as exposed
__all__ += exposed
from .LazyTensors import *; from .LazyTensors import __all__ as exposed
__all__ += exposed
from .Symbolic import *; from .Symbolic import __all__ as exposed
__all__ += exposed
from .Polynomials import *; from .Polynomials import __all__ as exposed
__all__ += exposed


from ..LazyTensors import Tensor
from .Derivatives import FiniteDifferenceDerivative
from ...Coordinerds import CoordinateSet, CoordinateSystem
from ..Symbolic import TensorDerivativeConverter
from ...Numputils import vec_tensordot
import numpy as np, itertools, copy, math

__all__ = [
    'FunctionExpansion'
]

########################################################################################################################
#
#                                           Polynomial
#
class TaylorPoly:
    """
    A handler for dealing with multidimensional polynomials

    Can be used to handle multiple polynomials simultaneously, but requires
    that all expansions are provided up to the same order.
    """

    def __init__(self,
                 derivatives,
                 transforms=None,
                 center=None,
                 ref=0,
                 weight_coefficients=False
                 ):
        """
        :param derivatives: Derivatives of the function being expanded
        :type derivatives: Iterable[np.ndarray | Tensor]
        :param transforms: Jacobian and higher order derivatives in the coordinates
        :type transforms: Iterable[np.ndarray | Tensor] | None
        :param center: the reference point for expanding aobut
        :type center: np.ndarray | None
        :param ref: the reference point value for shifting the expansion
        :type ref: float | np.ndarray
        :param weight_coefficients: whether the derivative terms need to be weighted or not
        :type weight_coefficients: bool
        """

        # raise NotImplementedError("doesn't deal with higher-order expansions properly yet")
        self._derivs = self.FunctionDerivatives(derivatives, weight_coefficients)
        self._center = np.asanyarray(center) if center is not None else center
        self.ref = np.asanyarray(ref) if not isinstance(ref, (int, float, np.integer, np.floating)) else ref
        if transforms is None:
            self._transf = None
        else:
            # transformation matrices from cartesians to internals
            self._transf = self.CoordinateTransforms(transforms)
        self._tensors = None
    @property
    def center(self):
        return self._center
    @property
    def is_multi(self):
        return self._derivs[0].ndim > 1
    @classmethod
    def multipolynomial(cls, *expansions):
        return cls(
            list(zip(*[e.expansion_tensors for e in expansions])),
            center=np.asanyarray([e.center for e in expansions]),
            ref=np.asanyarray([e.ref for e in expansions]),
            weight_coefficients=False
        )

    @property
    def expansion_tensors(self):
        """
        Provides the tensors that will contracted

        :return:
        :rtype: Iterable[np.ndarray]
        """
        if self._tensors is None:
            if self._transf is None:
                self._tensors = self._derivs
            else:
                self._tensors = TensorDerivativeConverter(self._transf, self._derivs).convert()
        return self._tensors
    @expansion_tensors.setter
    def expansion_tensors(self, tensors):
        """
        Are we going to assume in setting the tensors that
        people have done any requisite coordinate transformations?

        -> I think that's leaving room open for nasty surprises since we have
        the `_transf_ member
        -> but the `expansion_tensors` are in general expected to ... be transformed?

        Which means that we need _two_ properties 1) these tensors and 2) the derivs
        neither of which is transformed when being set
        """
        self._tensors = tensors

    @property
    def derivative_tensors(self):
        """
        Provides the base derivative tensors
        independent of any coordinate transformations

        :return:
        :rtype: Iterable[np.ndarray]
        """
        return self._derivs
    @derivative_tensors.setter
    def derivative_tensors(self, derivs):
        self._tensors = derivs

    def get_expansions(self, coords, subexpansions=None, outer=True, squeeze=True):
        """

        :param coords: Coordinates to evaluate the expansion at
        :type coords: np.ndarray | CoordinateSet
        :param subexpansions: A set of tensor expansion indices to use if we have a multiexpansion
        but only want some subset of the relevant points
        :type subexpansions: int | Iterable[int]
        :return:
        :rtype:
        """

        if not outer and not self.is_multi:
            raise ValueError("outer doesn't make sense without multiexpansion")
        outer = (not self.is_multi) or outer
        # doesn't really make sense if we don't have stuff to take the outer product over

        coords = np.asanyarray(coords)
        cshape = coords.shape # so we can squeeze at the end if need be
        if coords.ndim == 1:
            coords = coords[np.newaxis]
            outer = True # again...doesn't make sense _not_ to take an outer product here
        elif coords.ndim >= 2:
            if outer:
                if coords.ndim > 2: # we want to apply each expansion to each point
                    coords = np.reshape(coords, (-1, cshape[-1]))
            else:
                # we assume the first dimension is shared between the
                # multi-expansion and the coordinates
                coords = np.reshape(coords, (cshape[0], -1, cshape[-1]))

        center = self._center
        if center is None:
            # we assume we have a vector of points
            disp = coords
        else:
            # by default we take the outer product through
            # broacasting, but if outer==False I want to be
            # able to just thread things and require that the number
            # of vectors of points is the same as the numbers of expansions
            # which I think has to map onto the number of centers
            if outer:
                if center.ndim == 1:
                    center = center[np.newaxis]
                else:
                    center = center[np.newaxis, :, :]
                    coords = coords[:, np.newaxis]
            else:
                # we need to coerce things into the appropriate form
                # and by now we _know_ coords is rank three and now we just
                # need to make sure that center has the appropriate shape
                # for this
                if center.ndim == 1:
                    center = center[np.newaxis, np.newaxis]
                else:
                    center = center[:, np.newaxis, :] # need to broadcast along axis 1
                    # coords = coords[:, np.newaxis]

            disp = coords - center

        if isinstance(subexpansions, (int, np.integer)):
            subexpansions = [subexpansions]

        expansions=[]
        for i, tensr in enumerate(self.expansion_tensors):
            if subexpansions is not None:
                tensr = tensr[subexpansions] # I think it's really this simple...?
                # TBH I'm not entirely sure why one would want to use this kind of subexpansion
                # since it's generally not that expensive to build one of the smaller expansions
                # but it's nice to have the option since I can see this being a small optimization on the
                # edge somewhere.
            # contract the tensor by the displacements until it's completely reduced
            if outer:
                tensr = np.broadcast_to(tensr[np.newaxis], (disp.shape[0],) + tensr.shape)
            for j in range(i+1):
                # try:
                tensr = vec_tensordot(disp, tensr, axes=[[-1], [tensr.ndim-1]])
                # except:
                #     raise ValueError(disp.shape, tensr.shape, -1, tensr.ndim)
            contraction = tensr
            if squeeze:
                contraction = contraction.squeeze()
            expansions.append(contraction)

        if len(cshape) == 1:
            expansions = [np.squeeze(e) for e in expansions]
        elif len(cshape) > 2:
            new_shape = cshape[:-1]
            if outer:
                # need to reshape including the broadcast shape...
                # but at each order the contraction should be complete
                # meaning we lose the final dimension
                new_shape = (-1) + new_shape
            expansions = [np.reshape(e, new_shape) for e in expansions]


        return expansions
    def expand(self, coords, outer=True, squeeze=True):
        """Returns a numerical value for the expanded coordinates

        :param coords:
        :type coords: np.ndarray
        :return:
        :rtype: float | np.ndarray
        """
        ref = self.ref
        exps = self.get_expansions(coords, outer=outer, squeeze=squeeze)
        return ref + sum(exps)

    def __call__(self, coords, **kw):
        return self.expand(coords, **kw)

    class CoordinateTransforms:
        def __init__(self, transforms):
            self._transf = [np.asanyarray(t) for t in transforms]
        def __getitem__(self, i):
            if len(self._transf) < i:
                raise FunctionExpansionException("{}: transformations requested up to order {} but only provided to order {}".format(
                    type(self).__name__,
                    i,
                    len(self._transf)
                ))
            return self._transf[i]
        def __len__(self):
            return len(self._transf)
    class FunctionDerivatives:
        def __init__(self, derivs, weight_coefficients=True):
            self.derivs = [ np.asanyarray(t) for t in derivs ]
            if weight_coefficients:
                self.derivs = [self.weight_derivs(t, o+1) if weight_coefficients else t for o, t in enumerate(self.derivs)]
        def weight_derivs(self, t, order = None):
            """

            :param order:
            :type order: int
            :param t:
            :type t: Tensor
            :return:
            :rtype:
            """
            weighted = t
            if order is None:
                order = len(t.shape)
            if order > 1:
                weighted = weighted * (1/math.factorial(order))
                # s = t.shape
                # weights = np.ones(s)
                # all_inds = list(range(len(s)))
                # for i in range(2, order+1):
                #     for inds in itertools.combinations(all_inds, i):
                #         # define a diagonal slice through
                #         sel = tuple(slice(None, None, None) if a not in inds else np.arange(s[a]) for a in all_inds)
                #         weights[sel] = 1/math.factorial(i)
                # weighted = weighted.mul(weights)
                # print(weights, weighted.array)

            return weighted

        def __getitem__(self, i):
            if len(self.derivs) < i:
                raise FunctionExpansionException("{}: derivatives requested up to order {} but only provided to order {}".format(
                    type(self).__name__,
                    i,
                    len(self.derivs)
                ))
            return self.derivs[i]
        def __len__(self):
            return len(self.derivs)

    def deriv(self, which=None):
        """
        Computes the derivative(s) of the expansion(s) with respect to the
        supplied coordinates (`which=None` means compute the gradient)

        :param which:
        :type which:
        :return:
        :rtype:
        """

        expansions = self.expansion_tensors
        smol = isinstance(which, int)
        if smol:
            which = [which]
        elif which is None:
            which = list(range(expansions[0].shape[-1]))

        res = []
        cls = type(self)
        for i in which:
            derivs = []
            for j,e in enumerate(expansions):
                # need to know the shift dims?.... ?
                base_slice = tuple(slice(None, None, None) for _ in range(j+1))
                red = 0
                for k in range(j+1):
                    s = (...,) + base_slice[:k] + (i,) + base_slice[k+1:]
                    red = red + e[s] # this iteration is how we get the different multiples...?
                    # and even though it seems like it makes off diagonal mat els contribute too much
                    # this is actually correct!
                derivs.append(red)
            res.append(cls(
                derivs[1:],
                ref=derivs[0],
                center=self._center,
                weight_coefficients=False
            ))
        return cls.multipolynomial(*res)

    @classmethod
    def from_indices(cls, inds, ref=0, expansion_order=None, ndim=None, symmetrize=True, **opts):
        #TODO: handle via `SparseArray` methods
        if isinstance(inds, dict):
            inds = tuple(inds.items())
        ref = ref
        if ndim is None or expansion_order is None:
            if ndim is None:
                ndim = 0
            if expansion_order is None:
                expansion_order = 0
            for (i, v) in inds: # prescan to get tensor dimensions
                if len(i) == 0:
                    continue
                expansion_order = max(expansion_order, len(i))
                ndim = max(ndim, max(i)+1)
        t = [np.zeros((ndim,)*n) for n in range(1, expansion_order+1)]
        for (i, v) in inds:
            if len(i) == 0:
                ref = v
                continue
            tens = t[len(i)-1]
            if symmetrize:
                for p in itertools.permutations(i):
                    tens[p] = v
            else:
                tens[i] = v
        return cls(
            t,
            ref=ref,
            **opts
        )

    def _build_binomial_inds(self, g_partitions, w_partitions):
        """
        An unnecessarily efficient algorithm to get the set
        pairs of indices into the integer partitions of `g` and `w`
        plus an optional permutation of `g` which will lead to non-zero
        terms when the product of the pairwise binomial terms is computed


        :param g_partitions:
        :type g_partitions:
        :param w_partitions:
        :type w_partitions:
        :return:
        :rtype:
        """

        w_start = 0
        w_lens = np.count_nonzero(w_partitions, axis=1)
        g_lens = np.count_nonzero(g_partitions, axis=1)

        # build a mask telling us which leading terms are greater than
        # or equal to a given number
        len_mask = [
            np.arange(len(w_partitions))
        ]
        for i in range(1, max(w_lens)+1):
            cur = len_mask[-1]
            new = cur[w_lens[cur] >= i]
            len_mask.append(new)

        permutation_breakpoint = w_partitions[0]//2 + w_partitions[0] % 2
        partition_perm_inds = []
        for i, g in enumerate(g_partitions):
            # need to track _where_ we start in the w_partitions
            ng = g_lens[i]
            w_choice = w_partitions[len_mask[ng]]
            for j, w in enumerate(w_choice):
                if w[0] < g[0]: # I should be able to build a first-element mask
                    break
                if all(gg<=ww for gg,ww in zip(g[1:ng],w[1:ng])): # It's an open question if this can be faster?
                    partition_perm_inds.append([(i, len_mask[ng][j]), None])
                    if w[0] <= permutation_breakpoint:
                        # need to test against permutatons to see if it'll work
                        # TODO: use a bell-ringer pattern but with quick aborts by
                        #       ordering conditions
                        base_perm = np.arange(ng)
                        shift_idx = 0
                        for p in range():
                            ...


    def shift(self, new_origin):
        """
        Uses binomial expansion to new polynomial centered at the `new_origin`

        :param new_origin:
        :type new_origin:
        :return:
        :rtype:
        """
        from ...Combinatorics import Binomial, IntegerPartitionPermutations

        raise NotImplementedError("shifted polys on hold for a moment")

        cur_tensors = self.expansion_tensors
        n = len(cur_tensors)
        new_tensors = [0]*(n+1)

        delta = self.center - new_origin #????
        delta_prime = np.array([
            np.concatenate([delta[:i], [1], delta[i+1:]])
            for i in range(len(delta))
        ]).T

        binomials = Binomial(n+1)
        delta_tensors = np.full((n+1, n+1), None, dtype=object)
        binomial_tensors = np.full((n+1, n+1), None, dtype=object)

        iparts = [IntegerPartitionPermutations(i, dim=len(delta)) if i > 0 else None for i in range(n+1)]
        for g in range(n+1):
            if g > 0:
                gidx = iparts[g].partitions

            for w in range(g, n+1):
                if g > 0:
                    widx = iparts[w].partitions
                    binomial_tensors[w, g] = ...
                    # outer(d[w, g], dp, expand_axes=[w, -1]) -> would otherwise be an outer + a moveaxis
                    delta_tensors[w, g] = np.expand_dims(
                            delta_tensors[w-1, g-1],
                            [-1, w]
                        ) * np.expand_dims(
                            delta_prime,
                            np.concatenate([
                                np.arange(w),
                                np.arange(w, (w-1)+(g-1))
                            ])
                        )

                    D = binomial_tensors[w][g] * delta_tensors[w][g]
                else:
                    delta_tensors[w, 0] = np.expand_dims(
                            delta_tensors[w - 1, 0],
                            [-1]
                        ) * np.expand_dims(
                            delta,
                            np.arange(w)
                        )
                    D = delta_tensors[w][g]

                D *= (-1)**(w-g)

                new_tensors[g] += np.tensordot(cur_tensors[w], D, axes=list(range(w)))








########################################################################################################################
#
#                                           FunctionExpansion
#
class FunctionExpansionException(Exception):
    pass
class FunctionExpansion(TaylorPoly):
    """
    Specifically for expanding functions
    """

    @classmethod
    def expand_function(cls,
                        f, point,
                        order=4,
                        basis=None,
                        function_shape=None,
                        transforms=None,
                        weight_coefficients=True,
                        **fd_options
                        ):
        """
        Expands a function about a point up to the given order

        :param f:
        :type f: function
        :param point:
        :type point: np.ndarray | CoordinateSet
        :param order:
        :type order: int
        :param basis:
        :type basis: None | CoordinateSystem
        :param fd_options:
        :type fd_options:
        :return:
        :rtype:
        """

        derivs = FiniteDifferenceDerivative(f, function_shape=function_shape)(point, **fd_options)
        ref = f(point)
        dts = [derivs.derivative_tensor(i) for i in range(1, order + 1)]
        if transforms is None:
            if basis is not None and isinstance(point, CoordinateSet):
                transforms = [point.jacobian(basis, order=i) for i in range(order)]
            else:
                transforms = None
        return cls(dts, center=point, ref=ref, transforms=transforms, weight_coefficients=weight_coefficients)





